"""
SHELX (.res) input/output

Read/write files in SHELX (.res) file format.

Format documented at http://shelx.uni-ac.gwdg.de/SHELX/

Written by Martin Uhren and Georg Schusteritsch.
Adapted for ASE by James Kermode.
"""

from __future__ import division

import glob
import re

from ase.atoms import Atoms
from ase.geometry import cellpar_to_cell, cell_to_cellpar
from ase.calculators.calculator import Calculator
from ase.calculators.singlepoint import SinglePointCalculator

__all__ = ['Res', 'read_res', 'write_res']


class Res(object):

    """
    Object for representing the data in a Res file.
    Most attributes can be set directly.

    Args:
        atoms (Atoms):  Atoms object.

    .. attribute:: atoms

        Associated Atoms object.

    .. attribute:: name

        The name of the structure.

    .. attribute:: pressure

        The external pressure.

    .. attribute:: energy

        The internal energy of the structure.

    .. attribute:: spacegroup

        The space group of the structure.

    .. attribute:: times_found

        The number of times the structure was found.
    """

    def __init__(self, atoms, name=None, pressure=None,
                 energy=None, spacegroup=None, times_found=None):
        self.atoms_ = atoms
        if name is None:
            name = atoms.info.get('name')
        if pressure is None:
            pressure = atoms.info.get('pressure')
        if spacegroup is None:
            spacegroup = atoms.info.get('spacegroup')
        if times_found is None:
            times_found = atoms.info.get('times_found')
        self.name = name
        self.pressure = pressure
        self.energy = energy
        self.spacegroup = spacegroup
        self.times_found = times_found

    @property
    def atoms(self):
        """
        Returns Atoms object associated with this Res.
        """
        return self.atoms_

    @staticmethod
    def from_file(filename):
        """
        Reads a Res from a file.

        Args:
            filename (str): File name containing Res data.

        Returns:
            Res object.
        """
        with open(filename, 'r') as f:
            return Res.from_string(f.read())

    @staticmethod
    def parse_title(line):
        info = dict()

        tokens = line.split()
        num_tokens = len(tokens)
        # 1 = Name
        if num_tokens <= 1:
            return info
        info['name'] = tokens[1]
        # 2 = Pressure
        if num_tokens <= 2:
            return info
        info['pressure'] = float(tokens[2])
        # 3 = Volume
        # 4 = Internal energy
        if num_tokens <= 4:
            return info
        info['energy'] = float(tokens[4])
        # 5 = Spin density, 6 - Abs spin density
        # 7 = Space group OR num atoms (new format ONLY)
        idx = 7
        if tokens[idx][0] != '(':
            idx += 1

        if num_tokens <= idx:
            return info
        info['spacegroup'] = tokens[idx][1:len(tokens[idx]) - 1]
        # idx + 1 = n, idx + 2 = -
        # idx + 3 = times found
        if num_tokens <= idx + 3:
            return info
        info['times_found'] = int(tokens[idx + 3])

        return info

    @staticmethod
    def from_string(data):
        """
        Reads a Res from a string.

        Args:
            data (str): string containing Res data.

        Returns:
            Res object.
        """
        abc = []
        ang = []
        sp = []
        coords = []
        info = dict()
        coord_patt = re.compile("""(\w+)\s+
                                   ([0-9]+)\s+
                                   ([0-9\-\.]+)\s+
                                   ([0-9\-\.]+)\s+
                                   ([0-9\-\.]+)\s+
                                   ([0-9\-\.]+)""", re.VERBOSE)
        lines = data.splitlines()
        line_no = 0
        while line_no < len(lines):
            line = lines[line_no]
            tokens = line.split()
            if tokens:
                if tokens[0] == 'TITL':
                    try:
                        info = Res.parse_title(line)
                    except (ValueError, IndexError):
                        info = dict()
                elif tokens[0] == 'CELL' and len(tokens) == 8:
                    abc = [float(tok) for tok in tokens[2:5]]
                    ang = [float(tok) for tok in tokens[5:8]]
                elif tokens[0] == 'SFAC':
                    for atom_line in lines[line_no:]:
                        if line.strip() == 'END':
                            break
                        else:
                            match = coord_patt.search(atom_line)
                            if match:
                                sp.append(match.group(1))  # 1-indexed
                                cs = match.groups()[2:5]
                                coords.append([float(c) for c in cs])
                        line_no += 1  # Make sure the global is updated
            line_no += 1

        return Res(Atoms(symbols=sp,
                         scaled_positions=coords,
                         cell=cellpar_to_cell(list(abc) + list(ang)),
                         pbc=True, info=info),
                   info.get('name'),
                   info.get('pressure'),
                   info.get('energy'),
                   info.get('spacegroup'),
                   info.get('times_found'))

    def get_string(self, significant_figures=6, write_info=False):
        """
        Returns a string to be written as a Res file.

        Args:
            significant_figures (int): No. of significant figures to
                output all quantities. Defaults to 6.

            write_info (bool): if True, format TITL line using key-value pairs
               from atoms.info in addition to attributes stored in Res object

        Returns:
            String representation of Res.
        """

        # Title line
        if write_info:
            info = self.atoms.info.copy()
            for attribute in ['name', 'pressure', 'energy',
                              'spacegroup', 'times_found']:
                if getattr(self, attribute) and attribute not in info:
                    info[attribute] = getattr(self, attribute)
            lines = ['TITL ' + ' '.join(['{0}={1}'.format(k, v)
                                         for (k, v) in info.items()])]
        else:
            lines = ['TITL ' + self.print_title()]

        # Cell
        abc_ang = cell_to_cellpar(self.atoms.get_cell())
        fmt = '{{0:.{0}f}}'.format(significant_figures)
        cell = ' '.join([fmt.format(a) for a in abc_ang])
        lines.append('CELL 1.0 ' + cell)

        # Latt
        lines.append('LATT -1')

        # Atoms
        symbols = self.atoms.get_chemical_symbols()
        species_types = []
        for symbol in symbols:
            if symbol not in species_types:
                species_types.append(symbol)
        lines.append('SFAC ' + ' '.join(species_types))

        fmt = '{{0}} {{1}} {{2:.{0}f}} {{3:.{0}f}} {{4:.{0}f}} 1.0'
        fmtstr = fmt.format(significant_figures)
        for symbol, coords in zip(symbols,
                                  self.atoms_.get_scaled_positions()):
            lines.append(
                fmtstr.format(symbol,
                              species_types.index(symbol) + 1,
                              coords[0],
                              coords[1],
                              coords[2]))
        lines.append('END')
        return '\n'.join(lines)

    def __str__(self):
        """
        String representation of Res file.
        """
        return self.get_string()

    def write_file(self, filename, **kwargs):
        """
        Writes Res to a file. The supported kwargs are the same as those for
        the Res.get_string method and are passed through directly.
        """
        with open(filename, 'w') as f:
            f.write(self.get_string(**kwargs) + '\n')

    def print_title(self):
        tokens = [self.name, self.pressure, self.atoms.get_volume(),
                  self.energy, 0.0, 0.0, len(self.atoms)]
        if self.spacegroup:
            tokens.append('(' + self.spacegroup + ')')
        else:
            tokens.append('(P1)')
        if self.times_found:
            tokens.append('n - ' + str(self.times_found))
        else:
            tokens.append('n - 1')

        return ' '.join([str(tok) for tok in tokens])


def read_res(filename, index=-1):
    """
    Read input in SHELX (.res) format

    Multiple frames are read if `filename` contains a wildcard character,
    e.g. `file_*.res`. `index` specifes which frames to retun: default is
    last frame only (index=-1).
    """
    images = []
    for fn in sorted(glob.glob(filename)):
        res = Res.from_file(fn)
        if res.energy:
            calc = SinglePointCalculator(res.atoms,
                                         energy=res.energy)
            res.atoms.set_calculator(calc)
        images.append(res.atoms)
    return images[index]


def write_res(filename, images, write_info=True,
              write_results=True, significant_figures=6):
    """
    Write output in SHELX (.res) format

    To write multiple images, include a % format string in filename,
    e.g. `file_%03d.res`.

    Optionally include contents of Atoms.info dictionary if `write_info`
    is True, and/or results from attached calculator if `write_results`
    is True (only energy results are supported).
    """

    if not isinstance(images, (list, tuple)):
        images = [images]

    if len(images) > 1 and '%' not in filename:
        raise RuntimeError('More than one Atoms provided but no %' +
                           ' format string found in filename')

    for i, atoms in enumerate(images):
        fn = filename
        if '%' in filename:
            fn = filename % i
        res = Res(atoms)
        if write_results:
            calculator = atoms.get_calculator()
            if (calculator is not None and
                    isinstance(calculator, Calculator)):
                energy = calculator.results.get('energy')
                if energy is not None:
                    res.energy = energy
        res.write_file(fn, write_info=write_info,
                       significant_figures=significant_figures)



"""
This is the implementation of the exciting I/O functions
The functions are called with read write using the format "exi"

The module depends on lxml  http://lxml.de
"""

import numpy as np

from ase.atoms import Atoms
from ase.units import Bohr


def read_exciting(fileobj, index=-1):
    """Reads structure from exiting xml file.
    
    Parameters
    ----------
    fileobj: file object
        File handle from which data should be read.
        
    Other parameters
    ----------------
    index: integer -1
        Not used in this implementation.
    """
    
    from lxml import etree as ET

    # Parse file into element tree
    doc = ET.parse(fileobj)
    root = doc.getroot()
    speciesnodes = root.find('structure').getiterator('species')
    symbols = []
    positions = []
    basevects = []
    atoms = None
    # Collect data from tree
    for speciesnode in speciesnodes:
        symbol = speciesnode.get('speciesfile').split('.')[0]
        natoms = speciesnode.getiterator('atom')
        for atom in natoms:
            x, y, z = atom.get('coord').split()
            positions.append([float(x), float(y), float(z)])
            symbols.append(symbol)
    # scale unit cell accorting to scaling attributes
    if doc.xpath('//crystal/@scale'):
        scale = float(str(doc.xpath('//crystal/@scale')[0]))
    else:
        scale = 1
        
    if doc.xpath('//crystal/@stretch'):
        a, b, c = doc.xpath('//crystal/@scale')[0].split()
        stretch = np.array([float(a), float(b), float(c)])
    else:
        stretch = np.array([1.0, 1.0, 1.0])
    basevectsn = doc.xpath('//basevect/text()')
    for basevect in basevectsn:
        x, y, z = basevect.split()
        basevects.append(np.array([float(x) * Bohr * stretch[0],
                                   float(y) * Bohr * stretch[1],
                                   float(z) * Bohr * stretch[2]
                                   ]) * scale)
    atoms = Atoms(symbols=symbols, cell=basevects)
 
    atoms.set_scaled_positions(positions)
    if 'molecule' in root.find('structure').attrib.keys():
        if root.find('structure').attrib['molecule']:
            atoms.set_pbc(False)
    else:
        atoms.set_pbc(True)
        
    return atoms


def write_exciting(filename, images):
    """writes exciting input structure in XML
    
    Parameters
    ----------
    filename : str
        Name of file to which data should be written.
    images : Atom Object or List of Atoms objects
        This function will write the first Atoms object to file.
    
    Returns
    -------
    """
    from lxml import etree as ET
    fileobj = open(filename, 'wb')
    root = atoms2etree(images)
    fileobj.write(ET.tostring(root, method='xml',
                              pretty_print=True,
                              xml_declaration=True))


def atoms2etree(images):
    """This function creates the XML DOM corresponding
     to the structure for use in write and calculator
    
    Parameters
    ----------
    
    images : Atom Object or List of Atoms objects
    
    Returns
    -------
    root : etree object
        Element tree of exciting input file containing the structure
    """
    if not isinstance(images, (list, tuple)):
        images = [images]

    from lxml import etree as ET
    root = ET.Element('input')
    root.set(
        '{http://www.w3.org/2001/XMLSchema-instance}noNamespaceSchemaLocation',
        'http://xml.exciting-code.org/excitinginput.xsd')
     
    title = ET.SubElement(root, 'title')
    title.text = ''
    structure = ET.SubElement(root, 'structure')
    crystal = ET.SubElement(structure, 'crystal')
    atoms = images[0]
    for vec in atoms.cell:
        basevect = ET.SubElement(crystal, 'basevect')
        basevect.text = '%.14f %.14f %.14f' % tuple(vec / Bohr)
                            
    oldsymbol = ''
    oldrmt = -1
    newrmt = -1
    scaled = atoms.get_scaled_positions()
    for aindex, symbol in enumerate(atoms.get_chemical_symbols()):
        if 'rmt' in atoms.arrays:
            newrmt = atoms.get_array('rmt')[aindex] / Bohr
        if symbol != oldsymbol or newrmt != oldrmt:
            speciesnode = ET.SubElement(structure, 'species',
                                        speciesfile='%s.xml' % symbol,
                                        chemicalSymbol=symbol)
            oldsymbol = symbol
            if 'rmt' in atoms.arrays:
                oldrmt = atoms.get_array('rmt')[aindex] / Bohr
                if oldrmt > 0:
                    speciesnode.attrib['rmt'] = '%.4f' % oldrmt
           
        atom = ET.SubElement(speciesnode, 'atom',
                             coord='%.14f %.14f %.14f' % tuple(scaled[aindex]))
        if 'momenta' in atoms.arrays:
            atom.attrib['bfcmt'] = '%.14f %.14f %.14f' % tuple(
                atoms.get_array('mommenta')[aindex])
        
    return root



from ase import Atoms
from ase.units import fs, kB
from ase.calculators.test import TestPotential
from ase.md import Langevin
from ase.io import Trajectory, read
from ase.optimize import QuasiNewton
from ase.utils import seterr

with seterr(all='raise'):
    a = Atoms('4X',
              masses=[1, 2, 3, 4],
              positions=[(0, 0, 0),
                         (1, 0, 0),
                         (0, 1, 0),
                         (0.1, 0.2, 0.7)],
              calculator=TestPotential())
    print(a.get_forces())
    # Langevin should reproduce Verlet if friction is 0.
    md = Langevin(a, 0.5 * fs, 300 * kB, 0.0, logfile='-', loginterval=500)
    traj = Trajectory('4N.traj', 'w', a)
    md.attach(traj, 100)
    e0 = a.get_total_energy()
    md.run(steps=10000)
    del traj
    assert abs(read('4N.traj').get_total_energy() - e0) < 0.0001

    # Try again with nonzero friction.
    md = Langevin(a, 0.5 * fs, 300 * kB, 0.001, logfile='-', loginterval=500)
    traj = Trajectory('4NA.traj', 'w', a)
    md.attach(traj, 100)
    md.run(steps=10000)

    # We cannot test the temperature without a lot of statistics.
    # Asap does that.  But if temperature is quite unreasonable,
    # something is very wrong.
    T = a.get_temperature()
    assert T > 50
    assert T < 1000

    qn = QuasiNewton(a)
    qn.run(0.001)
    assert abs(a.get_potential_energy() - 1.0) < 0.000002



import numpy as np

from ase.build import bulk
from ase.calculators.test import gradient_test
from ase.calculators.lj import LennardJones
from ase.constraints import UnitCellFilter

a0 = bulk('Cu', cubic=True)

# perturb the atoms
s = a0.get_scaled_positions()
s[:, 0] *= 0.995
a0.set_scaled_positions(s)

# perturb the cell
a0.cell[...] += np.random.uniform(-1e-2, 1e-2,
                                  size=9).reshape((3,3))

atoms = a0.copy()
atoms.set_calculator(LennardJones())
ucf = UnitCellFilter(atoms)

# test all deritatives
f, fn = gradient_test(ucf)
assert abs(f - fn).max() < 1e-6



from distutils.core import setup
from glob import glob
from os.path import join

import os
import sys

long_description = """\
Jacapo is a python package providing an interface to Dacapo that
is compatible with the open source Atomic Simulation
Environment in the python scripting language."""

if sys.version_info < (2, 3, 0, 'final', 0):
    raise SystemExit('Python 2.3 or later is required!')

packages = ['Jacapo']

tools = ['tools/ncsum',
         'tools/plotnetcdf',
         'tools/pysub',
         'tools/qn_relax',
         'tools/stripnetcdf']

# Get the current version number:
exec(compile(open('version.py').read(), 'version.py', 'exec'))

setup(name = 'python-Jacapo',
      version=version,
      description='Jacapo - ase + Dacapo',
      url='http://www.fysik.dtu.dk/Campos/ase',
      maintainer='John Kitchin',
      maintainer_email='jkitchin@andrew.cmu.edu',
      license='LGPL',
      platforms=['linux'],
      packages=packages,
      scripts=tools,
      long_description=long_description)



from __future__ import print_function
import optparse

import numpy as np

from ase.data import covalent_radii
from ase.io.cube import read_cube_data
from ase.data.colors import cpk_colors
from ase.calculators.calculator import get_calculator


def plot(atoms, data, contours):
    """Plot atoms, unit-cell and iso-surfaces using Mayavi.
    
    Parameters:
        
    atoms: Atoms object
        Positions, atomiz numbers and unit-cell.
    data: 3-d ndarray of float
        Data for iso-surfaces.
    countours: list of float
        Contour values.
    """
    
    # Delay slow imports:
    from mayavi import mlab

    mlab.figure(1, bgcolor=(1, 1, 1))  # make a white figure

    # Plot the atoms as spheres:
    for pos, Z in zip(atoms.positions, atoms.numbers):
        mlab.points3d(*pos,
                      scale_factor=covalent_radii[Z],
                      resolution=20,
                      color=tuple(cpk_colors[Z]))

    # Draw the unit cell:
    A = atoms.cell
    for i1, a in enumerate(A):
        i2 = (i1 + 1) % 3
        i3 = (i1 + 2) % 3
        for b in [np.zeros(3), A[i2]]:
            for c in [np.zeros(3), A[i3]]:
                p1 = b + c
                p2 = p1 + a
                mlab.plot3d([p1[0], p2[0]],
                            [p1[1], p2[1]],
                            [p1[2], p2[2]],
                            tube_radius=0.1)

    cp = mlab.contour3d(data, contours=contours, transparent=True,
                        opacity=0.5, colormap='hot')
    # Do some tvtk magic in order to allow for non-orthogonal unit cells:
    polydata = cp.actor.actors[0].mapper.input
    pts = np.array(polydata.points) - 1
    # Transform the points to the unit cell:
    polydata.points = np.dot(pts, A / np.array(data.shape)[:, np.newaxis])
    
    # Apparently we need this to redraw the figure, maybe it can be done in
    # another way?
    mlab.view(azimuth=155, elevation=70, distance='auto')
    # Show the 3d plot:
    mlab.show()


description = """\
Plot iso-surfaces from a cube-file or a wave function or an electron
density from a calculator-restart file."""


def main(args=None):
    parser = optparse.OptionParser(usage='%prog [options] filename',
                                   description=description)
    add = parser.add_option
    add('-n', '--band-index', type=int, metavar='INDEX',
        help='Band index counting from zero.')
    add('-s', '--spin-index', type=int, metavar='SPIN',
        help='Spin index: zero or one.')
    add('-c', '--contours', default='4',
        help='Use "-c 3" for 3 contours or "-c -0.5,0.5" for specific ' +
        'values.  Default is four contours.')
    add('-r', '--repeat', help='Example: "-r 2,2,2".')
    add('-C', '--calculator-name', metavar='NAME', help='Name of calculator.')
    
    opts, args = parser.parse_args(args)
    if len(args) != 1:
        parser.error('Incorrect number of arguments')
        
    arg = args[0]
    if arg.endswith('.cube'):
        data, atoms = read_cube_data(arg)
    else:
        calc = get_calculator(opts.calculator_name)(arg, txt=None)
        atoms = calc.get_atoms()
        if opts.band_index is None:
            data = calc.get_pseudo_density(opts.spin_index)
        else:
            data = calc.get_pseudo_wave_function(opts.band_index,
                                                 opts.spin_index or 0)
            if data.dtype == complex:
                data = abs(data)
                
    mn = data.min()
    mx = data.max()
    print('Min: %16.6f' % mn)
    print('Max: %16.6f' % mx)
    
    if opts.contours.isdigit():
        n = int(opts.contours)
        d = (mx - mn) / n
        contours = np.linspace(mn + d / 2, mx - d / 2, n).tolist()
    else:
        contours = [float(x) for x in opts.contours.split(',')]
        
    if len(contours) == 1:
        print('1 contour:', contours[0])
    else:
        print('%d contours: %.6f, ..., %.6f' %
              (len(contours), contours[0], contours[-1]))

    if opts.repeat:
        repeat = [int(r) for r in opts.repeat.split(',')]
        data = np.tile(data, repeat)
        atoms *= repeat
        
    plot(atoms, data, contours)
    

if __name__ == '__main__':
    main()



"""Check if ase issues a warning if level shift breaks symmetry."""

from warnings import catch_warnings, simplefilter
from ase import Atoms
from ase.calculators.nwchem import NWChem


def main():
    """The main routine for the prove of the warning."""
    cr_atom = Atoms('Cr', positions=[(0, 0, 0)], pbc=False, magmoms=[5.0])
    calculator = NWChem(task='energy',
        geometry='nocenter noautosym noautoz',
        convergence={'energy': 1e-3,
                'density': 1e-2,
                'gradient': 5e-2},
        basis='Wachters+f',
        charge=1)
    cr_atom.set_calculator(calculator)
    with catch_warnings(record=True) as thrown_warning:
        simplefilter('always', RuntimeWarning)
        cr_atom.get_potential_energy()
        assert len(thrown_warning) == 1
        assert 'levelshift' in str(thrown_warning[-1].message)

if __name__ == '__main__':
    main()



# additional tests of the extended XYZ file I/O
# (which is also included in oi.py test case)
# maintainted by James Kermode <james.kermode@gmail.com>

import os

import numpy as np

import ase.io
from ase.atoms import Atoms
from ase.build import bulk

# array data of shape (N, 1) squeezed down to shape (N, ) -- bug fixed
# in commit r4541
at = bulk('Si')
ase.io.write('to.xyz', at, format='extxyz')
at.arrays['ns_extra_data'] = np.zeros((len(at), 1))
assert at.arrays['ns_extra_data'].shape == (2, 1)

ase.io.write('to_new.xyz', at, format='extxyz')
at_new = ase.io.read('to_new.xyz')
assert at_new.arrays['ns_extra_data'].shape == (2,)

os.unlink('to.xyz')
os.unlink('to_new.xyz')

# write sequence of images with different numbers of atoms -- bug fixed
# in commit r4542
images = [at, at * (2, 1, 1), at * (3, 1, 1)]
ase.io.write('multi.xyz', images, format='extxyz')
read_images = ase.io.read('multi.xyz@:')
assert read_images == images
os.unlink('multi.xyz')

# read xyz containing trailing blank line
f = open('structure.xyz', 'w')
f.write("""4
Coordinates
Mg        -4.25650        3.79180       -2.54123
C         -1.15405        2.86652       -1.26699
C         -5.53758        3.70936        0.63504
C         -7.28250        4.71303       -3.82016

""")
f.close()
a = ase.io.read('structure.xyz')
os.unlink('structure.xyz')

# read xyz with / in key value
f = open('slash.xyz', 'w')
f.write("""4
key1=a key2=a/b
Mg        -4.25650        3.79180       -2.54123
C         -1.15405        2.86652       -1.26699
C         -5.53758        3.70936        0.63504
C         -7.28250        4.71303       -3.82016
""")
f.close()
a = ase.io.read('slash.xyz')
assert a.info['key1'] == r'a'
assert a.info['key2'] == r'a/b'
os.unlink('slash.xyz')

struct = Atoms('H4', pbc=[True, True, True],
                cell=[[4.00759, 0.0, 0.0], [-2.003795, 3.47067475, 0.0], [3.06349683e-16, 5.30613216e-16, 5.00307]], positions=[[-2.003795e-05, 2.31379473, 0.875437189], [2.00381504, 1.15688001, 4.12763281], [2.00381504, 1.15688001, 3.37697219], [-2.003795e-05, 2.31379473, 1.62609781]])
struct.info = {'key_value_pairs': {'dataset': 'deltatest', 'kpoints': np.array([28, 28, 20]), 'identifier': 'deltatest_H_1.00'}, 'unique_id': '4cf83e2f89c795fb7eaf9662e77542c1'}

ase.io.write('tmp.xyz', struct)
os.unlink('tmp.xyz')



"""Function-like objects creating lattices with more than one element.

These lattice creators are mainly intended as examples for how to build you
own.  The following crystal structures are defined:

    B1 = NaCl = Rocksalt
    B2 = CsCl
    B3 = ZnS = Zincblende
    L1_2 = AuCu3
    L1_0 = AuCu
    TRI_Fe2O3
    HEX_Fe2O3
    
"""
from ase.lattice.cubic import DiamondFactory, SimpleCubicFactory
from ase.lattice.tetragonal import SimpleTetragonalFactory
from ase.lattice.triclinic import TriclinicFactory
from ase.lattice.hexagonal import HexagonalFactory


# To prevent a layer of element one on one side, and a layer of
# element two on the other side, NaCl is based on SimpleCubic instead
# of on FaceCenteredCubic
class NaClFactory(SimpleCubicFactory):
    "A factory for creating NaCl (B1, Rocksalt) lattices."

    bravais_basis = [[0, 0, 0], [0, 0, 0.5], [0, 0.5, 0], [0, 0.5, 0.5],
                     [0.5, 0, 0], [0.5, 0, 0.5], [0.5, 0.5, 0],
                     [0.5, 0.5, 0.5]]
    element_basis = (0, 1, 1, 0, 1, 0, 0, 1)
    

B1 = NaCl = Rocksalt = NaClFactory()


class CsClFactory(SimpleCubicFactory):
    "A factory for creating CsCl (B2) lattices."
    bravais_basis = [[0, 0, 0], [0.5, 0.5, 0.5]]
    element_basis = (0, 1)

B2 = CsCl = CsClFactory()


# The zincblende structure is easily derived from Diamond, which
# already has the right basis.
class ZnSFactory(DiamondFactory):
    "A factory for creating ZnS (B3, Zincblende) lattices."
    element_basis = (0, 1)

B3 = ZnS = Zincblende = ZnSFactory()


# The L1_0 structure is "based on FCC", but is a tetragonal distortion
# of fcc.  It must therefore be derived from the base-centered
# tetragonal structure.  That structure, however, does not exist,
# since it is equivalent to a simple tetragonal structure rotated 45
# degrees along the z-axis.  Basing L1_2 on that would however give
# unexpected miller indices.  L1_2 will therefore be based on a simple
# tetragonal structure, but with a basis corresponding to a
# base-centered tetragonal.
class AuCuFactory(SimpleTetragonalFactory):
    "A factory for creating AuCu (L1_0) lattices (tetragonal symmetry)."
    bravais_basis = [[0, 0, 0], [0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]
    element_basis = (0, 1, 1, 0)

AuCu = L1_0 = AuCuFactory()


# The L1_2 structure is "based on FCC", but is really simple cubic
# with a basis.
class AuCu3Factory(SimpleCubicFactory):
    "A factory for creating AuCu3 (L1_2) lattices."
    bravais_basis = [[0, 0, 0], [0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]
    element_basis = (0, 1, 1, 1)

AuCu3 = L1_2 = AuCu3Factory()


class TriclinicFe2O3Factory(TriclinicFactory):
    """A factory for creating hematite (Fe2O3) lattices.

     Rhombohedral unit cell.
     Pauling L, Hendricks S B
     Journal of the American Chemical Society 47 (1925) 781-790

     Example::

         #!/usr/bin/env python
    
         from ase.lattice.hexagonal import *
         from ase.lattice.compounds import *
         import ase.io as io
         from ase import Atoms, Atom
    
         index1=3
         index2=3
         index3=3
         mya = 5.42
         myb = 5.42
         myc = 5.42
         myalpha = 55.28
         mybeta = 55.28
         mygamma = 55.28
         gra = TRI_Fe2O3(symbol = ('Fe', 'O'),
                         latticeconstant={'a':mya,'b':myb, 'c':myc,
                                          'alpha':myalpha,
                                          'beta':mybeta,
                                          'gamma':mygamma},
                         size=(index1,index2,index3))
         io.write('rhombohedralUC_Fe2O3.xyz', gra, format='xyz')

     """

    bravais_basis = [[0.10534, 0.10534, 0.10534], [0.39466, 0.39466, 0.39466],
                     [0.60534, 0.60534, 0.60534], [0.89466, 0.89466, 0.89466],
                     [0.30569, 0.69431, 0.00000], [0.69431, 0.00000, 0.30569],
                     [0.00000, 0.30569, 0.69431], [0.19431, 0.80569, 0.50000],
                     [0.80569, 0.50000, 0.19431], [0.50000, 0.19431, 0.80569]]
    element_basis = (0, 0, 0, 0, 1, 1, 1, 1, 1, 1)

TRI_Fe2O3 = TriclinicFe2O3Factory()


class HexagonalFe2O3Factory(HexagonalFactory):
    """A factory for creating hematite (Fe2O3) lattices.
     With hexagonal unit cell.
     Blake R L, Hessevick R E, Zoltai T, Finger L W
     American Mineralogist 51 (1966) 123-129
     5.038 5.038 13.772 90 90 120 R-3c
     Fe       0 0 .3553  .0080  .0080 .00029  .0040      0      0
     O    .3059 0   1/4  .0068  .0083 .00046  .0042 .00058  .0012

     Example:
     #!/usr/bin/env python
     from ase.lattice.hexagonal import *
     from ase.lattice.compounds import *
     import ase.io as io
     from ase import Atoms, Atom

     index1=1
     index2=1
     index3=1
     mya = 5.038
     myb = 5.038
     myc = 13.772
     myalpha = 90
     mybeta = 90
     mygamma = 120
     gra = HEX_Fe2O3(symbol = ('Fe', 'O'),
     latticeconstant={'a':mya,'b':myb, 'c':myc,
     'alpha':myalpha,
     'beta':mybeta,
     'gamma':mygamma},
     size=(index1,index2,index3))
     io.write('hexaFe2O3.xyz', gra, format='xyz')
     
     """

    bravais_basis = [[0.000000, 0.000000, 0.355300],
                     [0.000000, 0.000000, 0.144700],
                     [0.000000, 0.000000, 0.644700],
                     [0.000000, 0.000000, 0.855300],
                     [0.666667, 0.333333, 0.688633],
                     [0.666667, 0.333333, 0.478033],
                     [0.666667, 0.333333, 0.978033],
                     [0.666667, 0.333333, 0.188633],
                     [0.333333, 0.666667, 0.021967],
                     [0.333333, 0.666667, 0.811367],
                     [0.333333, 0.666667, 0.311367],
                     [0.333333, 0.666667, 0.521967],
                     # Fe to O here
                     [0.305900, 0.000000, 0.250000],
                     [0.000000, 0.305900, 0.250000],
                     [0.694100, 0.694100, 0.250000],
                     [0.694100, 0.000000, 0.750000],
                     [0.000000, 0.694100, 0.750000],
                     [0.305900, 0.305900, 0.750000],
                     [0.972567, 0.333333, 0.583333],
                     [0.666667, 0.639233, 0.583333],
                     [0.360767, 0.027433, 0.583333],
                     [0.360767, 0.333333, 0.083333],
                     [0.666667, 0.027433, 0.083333],
                     [0.972567, 0.639233, 0.083333],
                     [0.639233, 0.666667, 0.916667],
                     [0.333333, 0.972567, 0.916667],
                     [0.027433, 0.360767, 0.916667],
                     [0.027433, 0.666667, 0.416667],
                     [0.333333, 0.360767, 0.416667],
                     [0.639233, 0.972567, 0.416667]]
    element_basis = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 1, 1, 1, 1, 1, 1, 1, 1,
                     1, 1, 1, 1, 1, 1, 1, 1, 1, 1)

HEX_Fe2O3 = HexagonalFe2O3Factory()



import os

from ase.test import NotAvailable
from ase.build import bulk
from ase.calculators.calculator import kpts2mp
from ase.calculators.elk import ELK

atoms = bulk('Al', 'bcc', a=4.0)

# save ELK_SPECIES_PATH
ELK_SPECIES_PATH = os.environ.get('ELK_SPECIES_PATH', None)
if ELK_SPECIES_PATH is None:
    raise NotAvailable('ELK_SPECIES_PATH not set.')

# find rmt of the default species
sfile = os.path.join(os.environ['ELK_SPECIES_PATH'], 'elk.in')
assert os.path.exists(sfile)
slines = open(sfile, 'r').readlines()
rmt_orig = {}
for name in ['Al']:
    found = False
    for n, line in enumerate(slines):
        if line.find("'" + name + "'") > -1:
            begline = n - 1
    for n, line in enumerate(slines[begline:]):
        if not line.strip():  # first empty line
            endline = n
            found = True
            break
    assert found
    # split needed because H is defined with comments
    rmt_orig[name] = float(slines[begline + 3].split()[0].strip())

assert rmt_orig['Al'] == 2.2  # 2.2 Bohr default

# test1

# generate species with custom rmt 2.1
rmt = {'Al': 2.1}
label = 'rmt2.1'

atomsrmt = atoms.copy()
os.environ['ELK_SPECIES_PATH'] = ELK_SPECIES_PATH
atomsrmt.calc = ELK(tasks=0, label=label, rmt=rmt)  # minimal calc
atomsrmt.get_potential_energy()
del atomsrmt.calc
del atomsrmt

# hack ELK_SPECIES_PATH to use custom species
os.environ['ELK_SPECIES_PATH'] = os.path.abspath(label) + '/'
# run calculation
calc = ELK(tasks=0, label=label,
           rgkmax=4.0, kpts=tuple(kpts2mp(atoms, 2.0, even=True)))
atoms.set_calculator(calc)
e1 = atoms.get_potential_energy()

# test2

# generate species with custom rmt 2.1
rmt = {'Al': -0.1}
label = 'rmt0.1m'

atomsrmt = atoms.copy()
os.environ['ELK_SPECIES_PATH'] = ELK_SPECIES_PATH
atomsrmt.calc = ELK(tasks=0, label=label, rmt=rmt)  # minimal calc
atomsrmt.get_potential_energy()
del atomsrmt.calc
del atomsrmt

# hack ELK_SPECIES_PATH to use custom species
os.environ['ELK_SPECIES_PATH'] = os.path.abspath(label) + '/'
# run calculation
calc = ELK(tasks=0, label=label,
           rgkmax=4.0, kpts=tuple(kpts2mp(atoms, 2.0, even=True)))
atoms.set_calculator(calc)
e2 = atoms.get_potential_energy()

# restore ELK_SPECIES_PATH
os.environ['ELK_SPECIES_PATH'] = ELK_SPECIES_PATH

assert abs(e1 - e2) < 1.0e-4



from ase.test import NotAvailable
from ase.build import bulk
from ase.dft.bandgap import get_band_gap
from ase.calculators.calculator import get_calculator

kpts = (4, 4, 4)

required = {'abinit': dict(ecut=200, toldfe=0.0001, chksymbreak=0),
            'aims': dict(sc_accuracy_rho=5.e-4,
                         output=['k_point_list',
                                 'k_eigenvalue,' +
                                 str(kpts[0] * kpts[1] * kpts[2])]),
            'elk': dict(tasks=0, rgkmax=5.0),
            'gpaw': dict(mode='pw')}


def run(name):
    Calculator = get_calculator(name)
    par = required.get(name, {})
    calc = Calculator(label=name + '_bandgap', xc='PBE',
                      # abinit, aims, elk - do not recognize the syntax below:
                      # https://trac.fysik.dtu.dk/projects/ase/ticket/98
                      # kpts={'size': kpts, 'gamma': True}, **par)
                      kpts=kpts, **par)
    si = bulk('Si', crystalstructure='diamond', a=5.43)
    si.calc = calc
    si.get_potential_energy()
    print(name, get_band_gap(si.calc))
    del si.calc
    # test spin-polarization
    calc = Calculator(label=name + '_bandgap_spinpol', xc='PBE',
                      # abinit, aims, elk - do not recognize the syntax below:
                      # https://trac.fysik.dtu.dk/projects/ase/ticket/98
                      # kpts={'size': kpts, 'gamma': True}, **par)
                      kpts=kpts, **par)
    si.set_initial_magnetic_moments([-0.1, 0.1])
    # this should not be necessary in the new ase interface standard ...
    if si.get_initial_magnetic_moments().any():  # spin-polarization
        if name == 'aims':
            calc.set(spin='collinear')
        if name == 'elk':
            calc.set(spinpol=True)
    si.set_calculator(calc)
    si.get_potential_energy()
    print(name, get_band_gap(si.calc))

# gpaw does not conform to the new ase interface standard:
# https://trac.fysik.dtu.dk/projects/gpaw/ticket/268
names = ['abinit', 'aims', 'elk']  # , 'gpaw']
for name in names:
    try:
        run(name)
    except NotAvailable:
        pass



from __future__ import print_function
from ase import Atoms
from ase.calculators.siesta import Siesta
from ase.units import Ry

a0 = 5.43
bulk = Atoms('Si2', [(0, 0, 0),
                     (0.25, 0.25, 0.25)],
             pbc=True)
b = a0 / 2
bulk.set_cell([(0, b, b),
               (b, 0, b),
               (b, b, 0)], scale_atoms=True)

calc = Siesta(label='Si',
              xc='PBE',
              mesh_cutoff=200 * Ry,
              energy_shift=0.01 * Ry,
              basis_set='DZ',
              kpts=[1, 2, 3],
              fdf_arguments={'DM.MixingWeight': 0.10,
                             'MaxSCFIterations': 10,
                             'DM.Tolerance': 0.1,
                             },
              )
bulk.set_calculator(calc)
e = bulk.get_potential_energy()
print(e)



from ase.test import cli, require
from ase.db import connect
from ase.io.jsonio import read_json
from ase.calculators.nwchem import NWChem

require('nwchem')
cli("""ase-build O | ase-run nwchem -d nwchem_cmdline.json &&
ase-build O2 | ase-run nwchem -d nwchem_cmdline.json""")
c = connect('nwchem_cmdline.json')
dct = read_json('nwchem_cmdline.json')
for name in ['O2', 'O']:
    d = c.get([('name', '=', name)])
    id = d.id
    e1 = d.energy
    e2 = c.get_atoms(id).get_potential_energy()
    e3 = NWChem.read_atoms(name).get_potential_energy()
    e4 = dct[id]['energy']
    assert e1 == e2 == e3 == e4
    print(e1)
ae = 2 * c.get('name=O').energy - c.get('name=O2').energy
assert abs(ae - 6.6053) < 1e-4



# -*- coding: utf-8 -*-

"""Test suit for the CP2K ASE calulator.

http://www.cp2k.org
Author: Ole Schuett <ole.schuett@mat.ethz.ch>
"""

from __future__ import division, print_function
import os

from ase.test import NotAvailable
from ase.build import molecule
from ase.calculators.cp2k import CP2K

inp = """
&FORCE_EVAL
   METHOD Quickstep
   &DFT
      BASIS_SET_FILE_NAME BASIS_MOLOPT
      &MGRID
         CUTOFF 400
      &END MGRID
      &XC
         &XC_FUNCTIONAL LDA
         &END XC_FUNCTIONAL
      &END XC
      &POISSON
         PERIODIC NONE
         PSOLVER  MT
      &END POISSON
   &END DFT
     &SUBSYS
      &KIND H
         BASIS_SET DZVP-MOLOPT-SR-GTH
         POTENTIAL GTH-LDA
      &END KIND
   &END SUBSYS
&END FORCE_EVAL
"""


def main():
    if "ASE_CP2K_COMMAND" not in os.environ:
        raise NotAvailable('$ASE_CP2K_COMMAND not defined')

    # Basically, the entire CP2K input is passed in explicitly.
    # Disable ASE's input generation by setting everything to None.
    # ASE should only add the CELL and the COORD section.
    calc = CP2K(basis_set=None,
                basis_set_file=None,
                max_scf=None,
                cutoff=None,
                force_eval_method=None,
                potential_file=None,
                poisson_solver=None,
                pseudo_potential=None,
                stress_tensor=False,
                xc=None,
                label='test_H2_inp', inp=inp)
    h2 = molecule('H2', calculator=calc)
    h2.center(vacuum=2.0)
    energy = h2.get_potential_energy()
    energy_ref = -30.6989595886
    diff = abs((energy - energy_ref) / energy_ref)
    assert diff < 1e-10
    print('passed test "H2_None"')


main()



from __future__ import print_function, division
# Copyright (C) 2010, Jesper Friis
# (see accompanying license files for details).

import numpy as np
from numpy import pi, sin, cos, arccos, sqrt, dot
from numpy.linalg import norm


def unit_vector(x):
    """Return a unit vector in the same direction as x."""
    y = np.array(x, dtype='float')
    return y / norm(y)


def angle(x, y):
    """Return the angle between vectors a and b in degrees."""
    return arccos(dot(x, y) / (norm(x) * norm(y))) * 180. / pi


def cell_to_cellpar(cell):
    """Returns the cell parameters [a, b, c, alpha, beta, gamma].

    Angles are in degrees.
    """
    lengths = [np.linalg.norm(v) for v in cell]
    angles = []
    for i in range(3):
        j = i - 1
        k = i - 2
        ll = lengths[j] * lengths[k]
        if ll > 1e-16:
            x = np.dot(cell[j], cell[k]) / ll
            angle = 180.0 / pi * arccos(x)
        else:
            angle = 90.0
        angles.append(angle)
    return np.array(lengths + angles)


def cellpar_to_cell(cellpar, ab_normal=(0, 0, 1), a_direction=None):
    """Return a 3x3 cell matrix from cellpar=[a,b,c,alpha,beta,gamma].

    Angles must be in degrees.

    The returned cell is orientated such that a and b
    are normal to `ab_normal` and a is parallel to the projection of
    `a_direction` in the a-b plane.

    Default `a_direction` is (1,0,0), unless this is parallel to
    `ab_normal`, in which case default `a_direction` is (0,0,1).

    The returned cell has the vectors va, vb and vc along the rows. The
    cell will be oriented such that va and vb are normal to `ab_normal`
    and va will be along the projection of `a_direction` onto the a-b
    plane.

    Example:

    >>> cell = cellpar_to_cell([1, 2, 4, 10, 20, 30], (0, 1, 1), (1, 2, 3))
    >>> np.round(cell, 3)
    array([[ 0.816, -0.408,  0.408],
           [ 1.992, -0.13 ,  0.13 ],
           [ 3.859, -0.745,  0.745]])

    """
    if a_direction is None:
        if np.linalg.norm(np.cross(ab_normal, (1, 0, 0))) < 1e-5:
            a_direction = (0, 0, 1)
        else:
            a_direction = (1, 0, 0)

    # Define rotated X,Y,Z-system, with Z along ab_normal and X along
    # the projection of a_direction onto the normal plane of Z.
    ad = np.array(a_direction)
    Z = unit_vector(ab_normal)
    X = unit_vector(ad - dot(ad, Z) * Z)
    Y = np.cross(Z, X)

    # Express va, vb and vc in the X,Y,Z-system
    alpha, beta, gamma = 90., 90., 90.
    if isinstance(cellpar, (int, float)):
        a = b = c = cellpar
    elif len(cellpar) == 1:
        a = b = c = cellpar[0]
    elif len(cellpar) == 3:
        a, b, c = cellpar
        alpha, beta, gamma = 90., 90., 90.
    else:
        a, b, c, alpha, beta, gamma = cellpar
    alpha *= pi / 180.0
    beta *= pi / 180.0
    gamma *= pi / 180.0
    va = a * np.array([1, 0, 0])
    vb = b * np.array([cos(gamma), sin(gamma), 0])
    cx = cos(beta)
    cy = (cos(alpha) - cos(beta) * cos(gamma)) / sin(gamma)
    cz = sqrt(1. - cx * cx - cy * cy)
    vc = c * np.array([cx, cy, cz])

    # Convert to the Cartesian x,y,z-system
    abc = np.vstack((va, vb, vc))
    T = np.vstack((X, Y, Z))
    cell = dot(abc, T)
    return cell


def metric_from_cell(cell):
    """Calculates the metric matrix from cell, which is given in the
    Cartesian system."""
    cell = np.asarray(cell, dtype=float)
    return np.dot(cell, cell.T)


def crystal_structure_from_cell(cell, eps=1e-4):
    """Return the crystal structure as a string calculated from the cell.

    Supply a cell (from atoms.get_cell()) and get a string representing
    the crystal structure returned. Works exactly the opposite
    way as ase.dft.kpoints.get_special_points().

    Parameters:

    cell : numpy.array or list
        An array like atoms.get_cell()

    Returns:

    crystal structure : str
        'cubic', 'fcc', 'bcc', 'tetragonal', 'orthorhombic',
        'hexagonal' or 'monoclinic'
    """
    cellpar = cell_to_cellpar(cell)
    abc = cellpar[:3]
    angles = cellpar[3:] / 180 * pi
    a, b, c = abc
    alpha, beta, gamma = angles
    if abc.ptp() < eps and abs(angles - pi / 2).max() < eps:
        return 'cubic'
    elif abc.ptp() < eps and abs(angles - pi / 3).max() < eps:
        return 'fcc'
    elif abc.ptp() < eps and abs(angles - np.arccos(-1 / 3)).max() < eps:
        return 'bcc'
    elif abs(a - b) < eps and abs(angles - pi / 2).max() < eps:
        return 'tetragonal'
    elif abs(angles - pi / 2).max() < eps:
        return 'orthorhombic'
    elif (abs(a - b) < eps and
          abs(gamma - pi / 3 * 2) < eps and
          abs(angles[:2] - pi / 2).max() < eps):
        return 'hexagonal'
    elif (c >= a and c >= b and alpha < pi / 2 and
          abs(angles[1:] - pi / 2).max() < eps):
        return 'monoclinic'
    else:
        raise ValueError('Cannot find crystal structure')


def complete_cell(cell):
    """Calculate complete cell with missing lattice vectors.

    Returns a new 3x3 ndarray.
    """

    cell = np.array(cell, dtype=float)
    missing = np.nonzero(~cell.any(axis=1))[0]

    if len(missing) == 3:
        cell.flat[::4] = 1.0
    if len(missing) == 2:
        # Must decide two vectors:
        i = 3 - missing.sum()
        assert abs(cell[i, missing]).max() < 1e-16, "Don't do that"
        cell[missing, missing] = 1.0
    elif len(missing) == 1:
        i = missing[0]
        cell[i] = np.cross(cell[i - 2], cell[i - 1])
        cell[i] /= np.linalg.norm(cell[i])

    return cell


def is_orthorhombic(cell):
    """Check that cell only has stuff in the diagonal."""
    return not (np.flatnonzero(cell) % 4).any()


def orthorhombic(cell):
    """Return cell as three box dimensions or raise ValueError."""
    if not is_orthorhombic(cell):
        raise ValueError('Not orthorhombic')
    return cell.diagonal().copy()



"""Mutation classes, that mutate the elements in the supplied
atoms objects."""
import random
import numpy as np

from ase.data import atomic_numbers
from ase.ga.offspring_creator import OffspringCreator


def chunks(l, n):
    """split a list into smaller chunks"""
    return [l[i:i + n] for i in range(0, len(l), n)]


class ElementMutation(OffspringCreator):
    """The base class for all operators where the elements
    of the atoms objects are mutated"""
    def __init__(self, element_pool, max_diff_elements,
                 min_percentage_elements, verbose, num_muts=1):
        OffspringCreator.__init__(self, verbose, num_muts=num_muts)
        if not isinstance(element_pool[0], (list, np.ndarray)):
            self.element_pools = [element_pool]
        else:
            self.element_pools = element_pool

        if max_diff_elements is None:
            self.max_diff_elements = [1e6 for _ in self.element_pools]
        elif isinstance(max_diff_elements, int):
            self.max_diff_elements = [max_diff_elements]
        else:
            self.max_diff_elements = max_diff_elements
        assert len(self.max_diff_elements) == len(self.element_pools)

        if min_percentage_elements is None:
            self.min_percentage_elements = [0 for _ in self.element_pools]
        elif isinstance(min_percentage_elements, (int, float)):
            self.min_percentage_elements = [min_percentage_elements]
        else:
            self.min_percentage_elements = min_percentage_elements
        assert len(self.min_percentage_elements) == len(self.element_pools)

        self.min_inputs = 1

    def get_new_individual(self, parents):
        raise NotImplementedError

    def get_mutation_index_list_and_choices(self, atoms):
        """Returns a list of the indices that are going to
        be mutated and a list of possible elements to mutate
        to. The lists obey the criteria set in the initialization.
        """
        itbm_ok = False
        while not itbm_ok:
            itbm = random.choice(range(len(atoms)))  # index to be mutated
            itbm_ok = True
            for i, e in enumerate(self.element_pools):
                if atoms[itbm].symbol in e:
                    elems = e[:]
                    elems_in, indices_in = zip(*[(a.symbol, a.index)
                                                 for a in atoms
                                                 if a.symbol in elems])
                    max_diff_elem = self.max_diff_elements[i]
                    min_percent_elem = self.min_percentage_elements[i]
                    if min_percent_elem == 0:
                        min_percent_elem = 1. / len(elems_in)
                    break
            else:
                itbm_ok = False

        # Check that itbm obeys min/max criteria
        diff_elems_in = len(set(elems_in))
        if diff_elems_in == max_diff_elem:
            # No more different elements allowed -> one element mutation
            ltbm = []  # list to be mutated
            for i in range(len(atoms)):
                if atoms[i].symbol == atoms[itbm].symbol:
                    ltbm.append(i)
        else:
            # Fewer or too many different elements already
            if self.verbose:
                print(int(min_percent_elem * len(elems_in)),
                      min_percent_elem, len(elems_in))
            all_chunks = chunks(indices_in,
                                int(min_percent_elem * len(elems_in)))
            itbm_num_of_elems = 0
            for a in atoms:
                if a.index == itbm:
                    break
                if a.symbol in elems:
                    itbm_num_of_elems += 1
            ltbm = all_chunks[itbm_num_of_elems //
                              (int(min_percent_elem * len(elems_in))) - 1]

        elems.remove(atoms[itbm].symbol)

        return ltbm, elems


class RandomElementMutation(ElementMutation):
    """Mutation that exchanges an element with a randomly chosen element from
    the supplied pool of elements
    If the individual consists of different groups of elements the element
    pool can be supplied as a list of lists

    Parameters:

    element_pool: List of elements in the phase space. The elements can be
        grouped if the individual consist of different types of elements.
        The list should then be a list of lists e.g. [[list1], [list2]]

    max_diff_elements: The maximum number of different elements in the
        individual. Default is infinite. If the elements are grouped
        max_diff_elements should be supplied as a list with each input
        corresponding to the elements specified in the same input in
        element_pool.

    min_percentage_elements: The minimum percentage of any element in the
        individual. Default is any number is allowed. If the elements are
        grouped min_percentage_elements should be supplied as a list with
        each input corresponding to the elements specified in the same input
        in element_pool.

    Example: element_pool=[[A,B,C,D],[x,y,z]], max_diff_elements=[3,2],
        min_percentage_elements=[.25, .5]
        An individual could be "D,B,B,C,x,x,x,x,z,z,z,z"
    """
    def __init__(self, element_pool, max_diff_elements=None,
                 min_percentage_elements=None, verbose=False,
                 num_muts=1):
        ElementMutation.__init__(self, element_pool, max_diff_elements,
                                 min_percentage_elements, verbose,
                                 num_muts=num_muts)
        self.descriptor = 'RandomElementMutation'

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        ltbm, choices = self.get_mutation_index_list_and_choices(f)

        new_element = random.choice(choices)
        for a in f:
            if a.index in ltbm:
                a.symbol = new_element
            indi.append(a)

        return (self.finalize_individual(indi),
                self.descriptor + ': Parent {0}'.format(f.info['confid']))


def mendeleiev_table():
    """Returns the mendeleiev table as a python list of lists.
    Each cell contains either None or a pair (symbol, atomic number),
    or a list of pairs for the cells \* and \**.
    """
    import re
    elems = 'HHeLiBeBCNOFNeNaMgAlSiPSClArKCaScTiVCrMnFeCoNiCuZnGaGeAsSeBrKrRb'
    elems += 'SrYZrNbMoTcRuRhPdAgCdInSnSbTeIXeCsBaLaCePrNdPmSmEuGdTbDyHoErTm'
    elems += 'YbLuHfTaWReOsIrPtAuHgTlPbBiPoAtRnFrRaAcThPaUNpPuAmCmBkCfEsFmMd'
    elems += 'NoLrRfDbSgBhHsMtDsRgUubUutUuqUupUuhUusUuo'
    L = [(e, i + 1)
         for (i, e) in enumerate(re.compile('[A-Z][a-z]*').findall(elems))]
    for i, j in ((88, 103), (56, 71)):
        L[i] = L[i:j]
        L[i + 1:] = L[j:]
    for i, j in ((12, 10), (4, 10), (1, 16)):
        L[i:i] = [None] * j
    return [L[18 * i:18 * (i + 1)] for i in range(7)]


def get_row_column(element):
    """Returns the row and column of the element in the periodic table.
    Note that Lanthanides and Actinides are defined to be group (column)
    3 elements"""
    t = mendeleiev_table()
    en = (element, atomic_numbers[element])
    for i in range(len(t)):
        for j in range(len(t[i])):
            if en == t[i][j]:
                return i, j
            elif isinstance(t[i][j], list):
                # Lanthanide or Actinide
                if en in t[i][j]:
                    return i, 3


class MoveDownMutation(ElementMutation):
    """
    Mutation that exchanges an element with an element one step
    (or more steps if fewer is forbidden) down the same
    column in the periodic table.

    This mutation is introduced and used in:
    P. B. Jensen et al., Phys. Chem. Chem. Phys., 16, 36, 19732-19740 (2014)

    The idea behind is that elements close to each other in the
    periodic table is chemically similar, and therefore exhibit
    similar properties. An individual in the population is
    typically close to fittest possible, exchanging an element
    with a similar element will normally result in a slight
    increase (or decrease) in fitness.

    Parameters:

    element_pool: List of elements in the phase space. The elements can be
        grouped if the individual consist of different types of elements.
        The list should then be a list of lists e.g. [[list1], [list2]]

    max_diff_elements: The maximum number of different elements in the
        individual. Default is infinite. If the elements are grouped
        max_diff_elements should be supplied as a list with each input
        corresponding to the elements specified in the same input in
        element_pool.

    min_percentage_elements: The minimum percentage of any element in the
        individual. Default is any number is allowed. If the elements are
        grouped min_percentage_elements should be supplied as a list with
        each input corresponding to the elements specified in the same input
        in element_pool.

    Example: element_pool=[[A,B,C,D],[x,y,z]], max_diff_elements=[3,2],
        min_percentage_elements=[.25, .5]
        An individual could be "D,B,B,C,x,x,x,x,z,z,z,z"
    """
    def __init__(self, element_pool, max_diff_elements=None,
                 min_percentage_elements=None, verbose=False,
                 num_muts=1):
        ElementMutation.__init__(self, element_pool, max_diff_elements,
                                 min_percentage_elements, verbose,
                                 num_muts=num_muts)
        self.descriptor = 'MoveDownMutation'

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        ltbm, choices = self.get_mutation_index_list_and_choices(f)
        # periodic table row, periodic table column
        ptrow, ptcol = get_row_column(f[ltbm[0]].symbol)

        popped = []
        m = 0
        for j in range(len(choices)):
            e = choices[j - m]
            row, column = get_row_column(e)
            if row <= ptrow or column != ptcol:
                # Throw away if above (lower numbered row)
                # or in a different column in the periodic table
                popped.append(choices.pop(j - m))
                m += 1

        used_descriptor = self.descriptor
        if len(choices) == 0:
            msg = '{0},{2} cannot be mutated by {1}, '
            msg = msg.format(f.info['confid'],
                             self.descriptor,
                             f[ltbm[0]].symbol)
            msg += 'doing random mutation instead'
            if self.verbose:
                print(msg)
            used_descriptor = 'RandomElementMutation_from_{0}'
            used_descriptor = used_descriptor.format(self.descriptor)
            random.shuffle(popped)
            choices = popped
        else:
            # Sorting the element that lie below and in the same column
            # in the periodic table so that the one closest below is first
            choices.sort(key=lambda x: get_row_column(x)[0])
        new_element = choices[0]

        for a in f:
            if a.index in ltbm:
                a.symbol = new_element
            indi.append(a)

        return (self.finalize_individual(indi),
                used_descriptor + ': Parent {0}'.format(f.info['confid']))


class MoveUpMutation(ElementMutation):
    """
    Mutation that exchanges an element with an element one step
    (or more steps if fewer is forbidden) up the same
    column in the periodic table.

    This mutation is introduced and used in:
    P. B. Jensen et al., Phys. Chem. Chem. Phys., 16, 36, 19732-19740 (2014)

    See MoveDownMutation for the idea behind

    Parameters:

    element_pool: List of elements in the phase space. The elements can be
        grouped if the individual consist of different types of elements.
        The list should then be a list of lists e.g. [[list1], [list2]]

    max_diff_elements: The maximum number of different elements in the
        individual. Default is infinite. If the elements are grouped
        max_diff_elements should be supplied as a list with each input
        corresponding to the elements specified in the same input in
        element_pool.

    min_percentage_elements: The minimum percentage of any element in the
        individual. Default is any number is allowed. If the elements are
        grouped min_percentage_elements should be supplied as a list with
        each input corresponding to the elements specified in the same input
        in element_pool.

    Example: element_pool=[[A,B,C,D],[x,y,z]], max_diff_elements=[3,2],
        min_percentage_elements=[.25, .5]
        An individual could be "D,B,B,C,x,x,x,x,z,z,z,z"
    """
    def __init__(self, element_pool, max_diff_elements=None,
                 min_percentage_elements=None, verbose=False, num_muts=1):
        ElementMutation.__init__(self, element_pool, max_diff_elements,
                                 min_percentage_elements, verbose,
                                 num_muts=num_muts)
        self.descriptor = 'MoveUpMutation'

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        ltbm, choices = self.get_mutation_index_list_and_choices(f)

        # periodic table row, periodic table column
        ptrow, ptcol = get_row_column(f[ltbm[0]].symbol)

        popped = []
        m = 0
        for j in range(len(choices)):
            e = choices[j - m]
            row, column = get_row_column(e)
            if row >= ptrow or column != ptcol:
                # Throw away if below (higher numbered row)
                # or in a different column in the periodic table
                popped.append(choices.pop(j - m))
                m += 1

        used_descriptor = self.descriptor
        if len(choices) == 0:
            msg = '{0},{2} cannot be mutated by {1}, '
            msg = msg.format(f.info['confid'],
                             self.descriptor,
                             f[ltbm[0]].symbol)
            msg += 'doing random mutation instead'
            if self.verbose:
                print(msg)
            used_descriptor = 'RandomElementMutation_from_{0}'
            used_descriptor = used_descriptor.format(self.descriptor)
            random.shuffle(popped)
            choices = popped
        else:
            # Sorting the element that lie above and in the same column
            # in the periodic table so that the one closest above is first
            choices.sort(key=lambda x: get_row_column(x)[0], reverse=True)
        new_element = choices[0]

        for a in f:
            if a.index in ltbm:
                a.symbol = new_element
            indi.append(a)

        return (self.finalize_individual(indi),
                used_descriptor + ': Parent {0}'.format(f.info['confid']))


class MoveRightMutation(ElementMutation):
    """
    Mutation that exchanges an element with an element one step
    (or more steps if fewer is forbidden) to the right in the
    same row in the periodic table.

    This mutation is introduced and used in:
    P. B. Jensen et al., Phys. Chem. Chem. Phys., 16, 36, 19732-19740 (2014)

    See MoveDownMutation for the idea behind

    Parameters:

    element_pool: List of elements in the phase space. The elements can be
        grouped if the individual consist of different types of elements.
        The list should then be a list of lists e.g. [[list1], [list2]]

    max_diff_elements: The maximum number of different elements in the
        individual. Default is infinite. If the elements are grouped
        max_diff_elements should be supplied as a list with each input
        corresponding to the elements specified in the same input in
        element_pool.

    min_percentage_elements: The minimum percentage of any element in the
        individual. Default is any number is allowed. If the elements are
        grouped min_percentage_elements should be supplied as a list with
        each input corresponding to the elements specified in the same input
        in element_pool.

    Example: element_pool=[[A,B,C,D],[x,y,z]], max_diff_elements=[3,2],
        min_percentage_elements=[.25, .5]
        An individual could be "D,B,B,C,x,x,x,x,z,z,z,z"
    """
    def __init__(self, element_pool, max_diff_elements=None,
                 min_percentage_elements=None, verbose=False, num_muts=1):
        ElementMutation.__init__(self, element_pool, max_diff_elements,
                                 min_percentage_elements, verbose,
                                 num_muts=num_muts)
        self.descriptor = 'MoveRightMutation'

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        ltbm, choices = self.get_mutation_index_list_and_choices(f)
        # periodic table row, periodic table column
        ptrow, ptcol = get_row_column(f[ltbm[0]].symbol)

        popped = []
        m = 0
        for j in range(len(choices)):
            e = choices[j - m]
            row, column = get_row_column(e)
            if row != ptrow or column <= ptcol:
                # Throw away if to the left (a lower numbered column)
                # or in a different row in the periodic table
                popped.append(choices.pop(j - m))
                m += 1

        used_descriptor = self.descriptor
        if len(choices) == 0:
            msg = '{0},{2} cannot be mutated by {1}, '
            msg = msg.format(f.info['confid'],
                             self.descriptor,
                             f[ltbm[0]].symbol)
            msg += 'doing random mutation instead'
            if self.verbose:
                print(msg)
            used_descriptor = 'RandomElementMutation_from_{0}'
            used_descriptor = used_descriptor.format(self.descriptor)
            random.shuffle(popped)
            choices = popped
        else:
            # Sorting so the element closest to the right is first
            choices.sort(key=lambda x: get_row_column(x)[1])
        new_element = choices[0]

        for a in f:
            if a.index in ltbm:
                a.symbol = new_element
            indi.append(a)

        return (self.finalize_individual(indi),
                used_descriptor + ': Parent {0}'.format(f.info['confid']))


class MoveLeftMutation(ElementMutation):
    """
    Mutation that exchanges an element with an element one step
    (or more steps if fewer is forbidden) to the left in the
    same row in the periodic table.

    This mutation is introduced and used in:
    P. B. Jensen et al., Phys. Chem. Chem. Phys., 16, 36, 19732-19740 (2014)

    See MoveDownMutation for the idea behind

    Parameters:

    element_pool: List of elements in the phase space. The elements can be
        grouped if the individual consist of different types of elements.
        The list should then be a list of lists e.g. [[list1], [list2]]

    max_diff_elements: The maximum number of different elements in the
        individual. Default is infinite. If the elements are grouped
        max_diff_elements should be supplied as a list with each input
        corresponding to the elements specified in the same input in
        element_pool.

    min_percentage_elements: The minimum percentage of any element in the
        individual. Default is any number is allowed. If the elements are
        grouped min_percentage_elements should be supplied as a list with
        each input corresponding to the elements specified in the same input
        in element_pool.

    Example: element_pool=[[A,B,C,D],[x,y,z]], max_diff_elements=[3,2],
        min_percentage_elements=[.25, .5]
        An individual could be "D,B,B,C,x,x,x,x,z,z,z,z"
    """
    def __init__(self, element_pool, max_diff_elements=None,
                 min_percentage_elements=None, verbose=False, num_muts=1):
        ElementMutation.__init__(self, element_pool, max_diff_elements,
                                 min_percentage_elements, verbose,
                                 num_muts=num_muts)
        self.descriptor = 'MoveLeftMutation'

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        ltbm, choices = self.get_mutation_index_list_and_choices(f)
        # periodic table row, periodic table column
        ptrow, ptcol = get_row_column(f[ltbm[0]].symbol)

        popped = []
        m = 0
        for j in range(len(choices)):
            e = choices[j - m]
            row, column = get_row_column(e)
            if row != ptrow or column >= ptcol:
                # Throw away if to the right (a higher numbered column)
                # or in a different row in the periodic table
                popped.append(choices.pop(j - m))
                m += 1

        used_descriptor = self.descriptor
        if len(choices) == 0:
            msg = '{0},{2} cannot be mutated by {1}, '
            msg = msg.format(f.info['confid'],
                             self.descriptor,
                             f[ltbm[0]].symbol)
            msg += 'doing random mutation instead'
            if self.verbose:
                print(msg)
            used_descriptor = 'RandomElementMutation_from_{0}'
            used_descriptor = used_descriptor.format(self.descriptor)
            random.shuffle(popped)
            choices = popped
        else:
            # Sorting so the element closest to the left is first
            choices.sort(key=lambda x: get_row_column(x)[1], reverse=True)
        new_element = choices[0]

        for a in f:
            if a.index in ltbm:
                a.symbol = new_element
            indi.append(a)

        return (self.finalize_individual(indi),
                used_descriptor + ':Parent {0}'.format(f.info['confid']))


class FullElementMutation(OffspringCreator):
    """Mutation that exchanges an all elements of a certain type with another
    randomly chosen element from the supplied pool of elements. Any constraints
    on the mutation are inhereted from the original candidate.

    Parameters:

    element_pool: List of elements in the phase space. The elements can be
        grouped if the individual consist of different types of elements.
        The list should then be a list of lists e.g. [[list1], [list2]]
    """
    def __init__(self, element_pool, verbose=False, num_muts=1):
        OffspringCreator.__init__(self, verbose, num_muts=num_muts)
        self.descriptor = 'FullElementMutation'
        if not isinstance(element_pool[0], (list, np.ndarray)):
            self.element_pools = [element_pool]
        else:
            self.element_pools = element_pool

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        # Randomly choose an element to mutate in the current individual.
        old_element = random.choice([a.symbol for a in f])
        # Find the list containing the chosen element. By choosing a new
        # element from the same list, the percentages are not altered.
        for i in range(len(self.element_pools)):
            if old_element in self.element_pools[i]:
                lm = i

        not_val = True
        while not_val:
            new_element = random.choice(self.element_pools[lm])
            not_val = new_element == old_element

        for a in f:
            if a.symbol == old_element:
                a.symbol = new_element
            indi.append(a)

        return (self.finalize_individual(indi),
                self.descriptor + ': Parent {0}'.format(f.info['confid']))



"""Berendsen NVT dynamics class."""

import numpy as np
from ase.md.md import MolecularDynamics
from ase.parallel import world


class NVTBerendsen(MolecularDynamics):
    """Berendsen (constant N, V, T) molecular dynamics.

    Usage: NVTBerendsen(atoms, timestep, temperature, taut, fixcm)

    atoms
        The list of atoms.
        
    timestep
        The time step.

    temperature
        The desired temperature, in Kelvin.

    taut
        Time constant for Berendsen temperature coupling.

    fixcm
        If True, the position and momentum of the center of mass is
        kept unperturbed.  Default: True.

    """

    def __init__(self, atoms, timestep, temperature, taut, fixcm=True,
                 trajectory=None, logfile=None, loginterval=1,
                 communicator=world):

        MolecularDynamics.__init__(self, atoms, timestep, trajectory,
                                   logfile, loginterval)
        self.taut = taut
        self.temperature = temperature
        self.fixcm = fixcm  # will the center of mass be held fixed?
        self.communicator = communicator

    def set_taut(self, taut):
        self.taut = taut

    def get_taut(self):
        return self.taut

    def set_temperature(self, temperature):
        self.temperature = temperature

    def get_temperature(self):
        return self.temperature

    def set_timestep(self, timestep):
        self.dt = timestep

    def get_timestep(self):
        return self.dt

    def scale_velocities(self):
        """ Do the NVT Berendsen velocity scaling """
        tautscl = self.dt / self.taut
        old_temperature = self.atoms.get_temperature()

        scl_temperature = np.sqrt(1.0 +
                                  (self.temperature / old_temperature - 1.0) *
                                  tautscl)
        # Limit the velocity scaling to reasonable values
        if scl_temperature > 1.1:
            scl_temperature = 1.1
        if scl_temperature < 0.9:
            scl_temperature = 0.9
        
        p = self.atoms.get_momenta()
        p = scl_temperature * p
        self.atoms.set_momenta(p)
        return

    def step(self, f):
        """Move one timestep forward using Berenden NVT molecular dynamics."""
        self.scale_velocities()

        # one step velocity verlet
        atoms = self.atoms
        p = self.atoms.get_momenta()
        p += 0.5 * self.dt * f

        if self.fixcm:
            # calculate the center of mass
            # momentum and subtract it
            psum = p.sum(axis=0) / float(len(p))
            p = p - psum

        self.atoms.set_positions(
            self.atoms.get_positions() +
            self.dt * p / self.atoms.get_masses()[:, np.newaxis])

        # We need to store the momenta on the atoms before calculating
        # the forces, as in a parallel Asap calculation atoms may
        # migrate during force calculations, and the momenta need to
        # migrate along with the atoms.  For the same reason, we
        # cannot use self.masses in the line above.

        self.atoms.set_momenta(p)
        f = self.atoms.get_forces()
        atoms.set_momenta(self.atoms.get_momenta() + 0.5 * self.dt * f)

        return f



"""Checks that writing and reading of NWChem input files is consistent."""

from ase.build import molecule
from ase.calculators.nwchem import NWChem
from ase import io

atoms = molecule('CH3COOH')

calc = NWChem()
calc.write_input(atoms)

atoms2 = io.read('nwchem.nw')

tol = 1e-8

check = sum(abs((atoms.positions - atoms2.positions).ravel()) > tol)
assert check == 0



import numpy as np

fast = False


def write_vti(filename, atoms, data=None):
    from vtk import vtkStructuredPoints, vtkDoubleArray, vtkXMLImageDataWriter

    #if isinstance(fileobj, str):
    #    fileobj = paropen(fileobj, 'w')
        
    if isinstance(atoms, list):
        if len(atoms) > 1:
            raise ValueError('Can only write one configuration to a VTI file!')
        atoms = atoms[0]

    if data is None:
        raise ValueError('VTK XML Image Data (VTI) format requires data!')

    data = np.asarray(data)

    if data.dtype == complex:
        data = np.abs(data)

    cell = atoms.get_cell()

    assert np.all(cell==np.diag(cell.diagonal())), 'Unit cell must be orthogonal'

    bbox = np.array(list(zip(np.zeros(3),cell.diagonal()))).ravel()

    # Create a VTK grid of structured points
    spts = vtkStructuredPoints()
    spts.SetWholeBoundingBox(bbox)
    spts.SetDimensions(data.shape)
    spts.SetSpacing(cell.diagonal() / data.shape)
    #spts.SetSpacing(paw.gd.h_c * Bohr)

    #print 'paw.gd.h_c * Bohr=',paw.gd.h_c * Bohr
    #print 'atoms.cell.diagonal() / data.shape=', cell.diagonal()/data.shape
    #assert np.all(paw.gd.h_c * Bohr==cell.diagonal()/data.shape)

    #s = paw.wfs.kpt_u[0].psit_nG[0].copy()
    #data = paw.get_pseudo_wave_function(band=0, kpt=0, spin=0, pad=False)
    #spts.point_data.scalars = data.swapaxes(0,2).flatten()
    #spts.point_data.scalars.name = 'scalars'

    # Allocate a VTK array of type double and copy data
    da = vtkDoubleArray()
    da.SetName('scalars')
    da.SetNumberOfComponents(1)
    da.SetNumberOfTuples(np.prod(data.shape))

    for i,d in enumerate(data.swapaxes(0,2).flatten()):
        da.SetTuple1(i,d)

    # Assign the VTK array as point data of the grid
    spd = spts.GetPointData() # type(spd) is vtkPointData
    spd.SetScalars(da)

    """
    from vtk.util.vtkImageImportFromArray import vtkImageImportFromArray
    iia = vtkImageImportFromArray()
    #iia.SetArray(Numeric_asarray(data.swapaxes(0,2).flatten()))
    iia.SetArray(Numeric_asarray(data))
    ida = iia.GetOutput()
    ipd = ida.GetPointData()
    ipd.SetName('scalars')
    spd.SetScalars(ipd.GetScalars())
    """

    # Save the ImageData dataset to a VTK XML file.
    w = vtkXMLImageDataWriter()

    if fast:
        w.SetDataModeToAppend()
        w.EncodeAppendedDataOff()
    else:
        w.SetDataModeToAscii()

    w.SetFileName(filename)
    w.SetInput(spts)
    w.Write()


def write_vtu(filename, atoms, data=None):
    from vtk import vtkUnstructuredGrid, vtkPoints, vtkXMLUnstructuredGridWriter
    #if isinstance(fileobj, str):
    #    fileobj = paropen(fileobj, 'w')
        
    if isinstance(atoms, list):
        if len(atoms) > 1:
            raise ValueError('Can only write one configuration to a VTI file!')
        atoms = atoms[0]

    """
    if data is None:
        raise ValueError('VTK XML Unstructured Grid (VTI) format requires data!')

    data = np.asarray(data)

    if data.dtype == complex:
        data = np.abs(data)
    """

    cell = atoms.get_cell()

    assert np.all(cell==np.diag(cell.diagonal())), 'Unit cell must be orthogonal' #TODO bounding box with general unit cell?!

    bbox = np.array(list(zip(np.zeros(3),cell.diagonal()))).ravel()

    # Create a VTK grid of structured points
    ugd = vtkUnstructuredGrid()
    ugd.SetWholeBoundingBox(bbox)

    """
    # Allocate a VTK array of type double and copy data
    da = vtkDoubleArray()
    da.SetName('scalars')
    da.SetNumberOfComponents(3)
    da.SetNumberOfTuples(len(atoms))

    for i,pos in enumerate(atoms.get_positions()):
        da.SetTuple3(i,pos[0],pos[1],pos[2])
    """
    p = vtkPoints()
    p.SetNumberOfPoints(len(atoms))
    p.SetDataTypeToDouble()
    for i,pos in enumerate(atoms.get_positions()):
        p.InsertPoint(i,pos[0],pos[1],pos[2])


    ugd.SetPoints(p)

    # Assign the VTK array as point data of the grid
    #upd = ugd.GetPointData() # type(spd) is vtkPointData
    #upd.SetScalars(da)


    # Save the UnstructuredGrid dataset to a VTK XML file.
    w = vtkXMLUnstructuredGridWriter()

    if fast:
        w.SetDataModeToAppend()
        w.EncodeAppendedDataOff()
    else:
        w.GetCompressor().SetCompressionLevel(0)
        w.SetDataModeToAscii()

    w.SetFileName(filename)
    w.SetInput(ugd)
    w.Write()



from ase import Atoms
a = Atoms('H2')

a[0].magmom = 1
m = a.get_initial_magnetic_moments()
assert m.shape == (2,) and (m == [1, 0]).all()

a[1].magmom = -1
m = a.get_initial_magnetic_moments()
assert m.shape == (2,) and (m == [1, -1]).all()
assert a[1].magmom == -1

a.set_initial_magnetic_moments()
a[0].magmom = (0, 1, 0)
m = a.get_initial_magnetic_moments()
assert m.shape == (2, 3) and (m == [(0, 1, 0), (0, 0, 0)]).all()

a[1].magmom = (1, 0, 0)
m = a.get_initial_magnetic_moments()
assert m.shape == (2, 3) and (m == [(0, 1, 0), (1, 0, 0)]).all()
assert (a[1].magmom == (1, 0, 0)).all()



from __future__ import print_function

import numpy as np
from ase import Atoms
from ase.io import write, read, iread
from ase.io.formats import all_formats, get_ioformat
from ase.calculators.singlepoint import SinglePointCalculator

try:
    import matplotlib
except ImportError:
    matplotlib = 0

try:
    from lxml import etree
except ImportError:
    etree = 0

try:
    import Scientific
except ImportError:
    Scientific = 0

a = 5.0
d = 1.9
c = a / 2
atoms = Atoms('AuH',
              positions=[(0, c, c), (d, c, c)],
              cell=(2 * d, a, a),
              pbc=(1, 0, 0))
extra = np.array([2.3, 4.2])
atoms.set_array('extra', extra)
atoms *= (2, 1, 1)

# attach some results to the Atoms. These are serialised by the extxyz writer.
spc = SinglePointCalculator(atoms,
                            energy=-1.0,
                            stress=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
                            forces=-1.0 * atoms.positions)
atoms.set_calculator(spc)
images = [atoms, atoms]


def check(a, format):
    assert abs(a.positions - atoms.positions).max() < 1e-6, (a.positions -
                                                             atoms.positions)
    if format in ['traj', 'cube', 'cfg', 'struct', 'gen', 'extxyz',
                  'db', 'json', 'trj']:
        assert abs(a.cell - atoms.cell).max() < 1e-6
    if format in ['cfg', 'extxyz']:
        assert abs(a.get_array('extra') -
                   atoms.get_array('extra')).max() < 1e-6
    if format in ['extxyz', 'traj', 'trj', 'db', 'json']:
        assert (a.pbc == atoms.pbc).all()
        assert a.get_potential_energy() == atoms.get_potential_energy()
        assert (a.get_stress() == atoms.get_stress()).all()
        assert abs(a.get_forces() - atoms.get_forces()).max() < 1e-12

for format in sorted(all_formats):
    if format in ['abinit', 'castep-cell', 'dftb', 'eon', 'gaussian']:
        # Someone should do something ...
        continue

    if format in ['v-sim']:
        # Standalone test used as not compatible with 1D periodicity
        continue

    if format in ['postgresql', 'trj', 'vti', 'vtu']:
        # Let's not worry about these.
        continue

    if not matplotlib and format in ['eps', 'png']:
        continue

    if not etree and format == 'exciting':
        continue

    if not Scientific and format == 'etsf':
        continue

    io = get_ioformat(format)
    print('{0:20}{1}{2}{3}{4}'.format(format,
                                      ' R'[bool(io.read)],
                                      ' W'[bool(io.write)],
                                      '+1'[io.single],
                                      'SF'[io.acceptsfd]))
    fname1 = 'io-test.1.' + format
    fname2 = 'io-test.2.' + format
    if io.write:
        write(fname1, atoms, format=format)
        if not io.single:
            write(fname2, images, format=format)

        if io.read:
            for a in [read(fname1, format=format), read(fname1)]:
                check(a, format)

            if not io.single:
                if format in ['json', 'db']:
                    aa = read(fname2 + '@id=1') + read(fname2 + '@id=2')
                else:
                    aa = [read(fname2), read(fname2, 0)]
                aa += read(fname2, ':')
                for a in iread(fname2, format=format):
                    aa.append(a)
                assert len(aa) == 6, aa
                for a in aa:
                    check(a, format)



from math import radians, sin, cos

from ase import Atoms
from ase.neb import NEB
from ase.constraints import FixAtoms
from ase.calculators.emt import EMT
from ase.optimize import QuasiNewton, BFGS
from ase.visualize import view

# http://jcp.aip.org/resource/1/jcpsa6/v97/i10/p7507_s1
doo = 2.74
doht = 0.957
doh = 0.977
angle = radians(104.5)
initial = Atoms('HOHOH',
                positions=[(-sin(angle) * doht, 0., cos(angle) * doht),
                           (0., 0., 0.),
                           (0., 0., doh),
                           (0., 0., doo),
                           (sin(angle) * doht, 0., doo - cos(angle) * doht)])
if 0:
    view(initial)

final = Atoms('HOHOH',
              positions=[(-sin(angle) * doht, 0., cos(angle) * doht),
                         (0., 0., 0.),
                         (0., 0., doo - doh),
                         (0., 0., doo),
                         (sin(angle) * doht, 0., doo - cos(angle) * doht)])
if 0:
    view(final)

# Make band:
images = [initial.copy()]
for i in range(3):
    images.append(initial.copy())
images.append(final.copy())
neb = NEB(images, climb=True)

# Set constraints and calculator:
constraint = FixAtoms(indices=[1, 3])  # fix OO
for image in images:
    image.set_calculator(EMT())
    image.set_constraint(constraint)

for image in images:  # O-H(shared) distance
    print(image.get_distance(1, 2), image.get_potential_energy())

# Relax initial and final states:
if 1:
    # XXX: Warning:
    # One would have to optimize more tightly in order to get
    # symmetric anion from both images[0] and [1], but
    # if one optimizes tightly one gets rotated(H2O) ... OH- instead
    dyn1 = QuasiNewton(images[0])
    dyn1.run(fmax=0.01)
    dyn2 = QuasiNewton(images[-1])
    dyn2.run(fmax=0.01)

# Interpolate positions between initial and final states:
neb.interpolate()

for image in images:
    print(image.get_distance(1, 2), image.get_potential_energy())

dyn = BFGS(neb, trajectory='emt_h3o2m.traj')
dyn.run(fmax=0.05)

for image in images:
    print(image.get_distance(1, 2), image.get_potential_energy())



# Copyright (C) 2012, Jesper Friis, SINTEF
# (see accompanying license files for ASE).
"""Module to read and write atoms EON reactant.con files.

See http://theory.cm.utexas.edu/eon/index.html for a description of EON.
"""
import os
from warnings import warn
from glob import glob

import numpy as np

from ase.atoms import Atoms
from ase.constraints import FixAtoms
from ase.geometry import cellpar_to_cell, cell_to_cellpar
from ase.parallel import paropen


def read_eon(fileobj):
    """Reads an EON reactant.con file.  If *fileobj* is the name of a
    "states" directory created by EON, all the structures will be read."""
    if isinstance(fileobj, str):
        if (os.path.isdir(fileobj)):
            return read_states(fileobj)
        else:
            f = open(fileobj)
    else:
        f = fileobj

    comment = f.readline().strip()
    f.readline()  # 0.0000 TIME  (??)
    cell_lengths = f.readline().split()
    cell_angles = f.readline().split()
    # Different order of angles in EON.
    cell_angles = [cell_angles[2], cell_angles[1], cell_angles[0]]
    cellpar = [float(x) for x in cell_lengths + cell_angles]
    f.readline()  # 0 0     (??)
    f.readline()  # 0 0 0   (??)
    ntypes = int(f.readline())  # number of atom types
    natoms = [int(n) for n in f.readline().split()]
    atommasses = [float(m) for m in f.readline().split()]

    symbols = []
    coords = []
    masses = []
    fixed = []
    for n in range(ntypes):
        symbol = f.readline().strip()
        symbols.extend([symbol] * natoms[n])
        masses.extend([atommasses[n]] * natoms[n])
        f.readline()  # Coordinates of Component n
        for i in range(natoms[n]):
            row = f.readline().split()
            coords.append([float(x) for x in row[:3]])
            fixed.append(bool(int(row[3])))

    if isinstance(fileobj, str):
        f.close()

    atoms = Atoms(symbols=symbols,
                  positions=coords,
                  masses=masses,
                  cell=cellpar_to_cell(cellpar),
                  constraint=FixAtoms(mask=fixed),
                  info=dict(comment=comment))
    
    return atoms
                  

def read_states(states_dir):
    """Read structures stored by EON in the states directory *states_dir*."""
    subdirs = glob(os.path.join(states_dir, '[0123456789]*'))
    subdirs.sort(key=lambda d: int(os.path.basename(d)))
    images = [read_eon(os.path.join(subdir, 'reactant.con'))
              for subdir in subdirs]
    return images

    
def write_eon(fileobj, images):
    """Writes structure to EON reactant.con file
    Multiple snapshots are not allowed."""
    if isinstance(fileobj, str):
        f = paropen(fileobj, 'w')
    else:
        f = fileobj

    if isinstance(images, Atoms):
        atoms = images
    elif len(images) == 1:
        atoms = images[0]
    else:
        raise ValueError('Can only write one configuration to EON '
                         'reactant.con file')

    out = []
    out.append(atoms.info.get('comment', 'Generated by ASE'))
    out.append('0.0000 TIME')  # ??

    a, b, c, alpha, beta, gamma = cell_to_cellpar(atoms.cell)
    out.append('%-10.6f  %-10.6f  %-10.6f' % (a, b, c))
    out.append('%-10.6f  %-10.6f  %-10.6f' % (gamma, beta, alpha))

    out.append('0 0')    # ??
    out.append('0 0 0')  # ??

    symbols = atoms.get_chemical_symbols()
    massdict = dict(list(zip(symbols, atoms.get_masses())))
    atomtypes = sorted(massdict.keys())
    atommasses = [massdict[at] for at in atomtypes]
    natoms = [symbols.count(at) for at in atomtypes]
    ntypes = len(atomtypes)

    out.append(str(ntypes))
    out.append(' '.join([str(n) for n in natoms]))
    out.append(' '.join([str(n) for n in atommasses]))

    atom_id = 0
    for n in range(ntypes):
        fixed = np.array([False] * len(atoms))
        out.append(atomtypes[n])
        out.append('Coordinates of Component %d' % (n + 1))
        indices = [i for i, at in enumerate(symbols) if at == atomtypes[n]]
        a = atoms[indices]
        coords = a.positions
        for c in a.constraints:
            if not isinstance(c, FixAtoms):
                warn('Only FixAtoms constraints are supported by con files. '
                     'Dropping %r', c)
                continue
            if c.index.dtype.kind == 'b':
                fixed = np.array(c.index, dtype=int)
            else:
                fixed = np.zeros((natoms[n], ), dtype=int)
                for i in c.index:
                    fixed[i] = 1
        for xyz, fix in zip(coords, fixed):
            out.append('%22.17f %22.17f %22.17f %d %4d' %
                       (tuple(xyz) + (fix, atom_id)))
            atom_id += 1
    f.write('\n'.join(out))
    f.write('\n')

    if isinstance(fileobj, str):
        f.close()



"""Test Atoms.__delitem__ with FixAtoms constraint."""
from ase import Atoms
from ase.constraints import FixAtoms

for i, j in [(slice(0, -1), None),
             (slice(0, 1), [0]),
             (slice(0, None), None),
             (0, [0]),
             (1, [0]),
             (2, [0, 1]),
             (-1, [0, 1])]:
    a = Atoms('H3')
    a.constraints = FixAtoms(indices=[0, 1])
    del a[i]
    print(i, j, a.constraints)
    if j is None:
        assert len(a.constraints) == 0
    else:
        assert (a.constraints[0].index == j).all()



from __future__ import division, print_function
import numpy as np
from ase.build import bulk
from ase.calculators.lj import LennardJones
from ase.constraints import UnitCellFilter
from ase.optimize import MDMin

# Theoretical infinite-cutoff LJ FCC unit cell parameters
vol0 = 4 * 0.91615977036  # theoretical minimum
a0 = vol0**(1 / 3)

a = bulk('X', 'fcc', a=a0)
cell0 = a.get_cell()

a.calc = LennardJones()
a.set_cell(np.dot(a.cell,
                  [[1.02, 0, 0.03],
                   [0, 0.99, -0.02],
                   [0.1, -0.01, 1.03]]),
           scale_atoms=True)

a *= (1, 2, 3)
cell0 *= np.array([1, 2, 3])[:, np.newaxis]

a.rattle()

# Verify analytical stress tensor against numerical value
s_analytical = a.get_stress()
s_numerical = a.calc.calculate_numerical_stress(a, 1e-5)
s_p_err = 100 * (s_numerical - s_analytical) / s_numerical

print("Analytical stress:\n", s_analytical)
print("Numerical stress:\n", s_numerical)
print("Percent error in stress:\n", s_p_err)
assert np.all(abs(s_p_err) < 1e-5)

# Minimize unit cell
opt = MDMin(UnitCellFilter(a), dt=0.01)
opt.run(fmax=1e-3)

# Verify minimized unit cell using Niggli tensors
g_minimized = np.dot(a.cell, a.cell.T)
g_theory = np.dot(cell0, cell0.T)
g_p_err = 100 * (g_minimized - g_theory) / g_theory

print("Minimized Niggli tensor:\n", g_minimized)
print("Theoretical Niggli tensor:\n", g_theory)
print("Percent error in Niggli tensor:\n", g_p_err)
assert np.all(abs(g_p_err) < 1)



import functools
import warnings
from ase.build.bulk import bulk as newbulk
__all__ = ['bulk']


@functools.wraps(newbulk)
def bulk(*args, **kwargs):
    warnings.warn('Use ase.build.bulk() instead', stacklevel=2)
    return newbulk(*args, **kwargs)



def installed():
    import os
    from ase.test import NotAvailable
    vcmd = os.getenv('VASP_COMMAND')
    vscr = os.getenv('VASP_SCRIPT')
    if vcmd == None and vscr == None:
        raise NotAvailable('Neither VASP_COMMAND nor VASP_SCRIPT defined')
    return True



import optparse
import sys

from ase.test import test

usage = ('Usage: python -m ase.test [-c calc1,calc2,...] '
         '[test1.py test2.py ...]')
parser = optparse.OptionParser(usage=usage, description='Test ASE')

parser.add_option('-c', '--calculators',
                  help='Comma-separated list of calculators to test.')
parser.add_option('-v', '--verbosity', type=int, default=2, metavar='N',
                  help='Use 0, 1 or 2.')

opts, args = parser.parse_args()

if opts.calculators:
    calculators = opts.calculators.split(',')
else:
    calculators = []

results = test(verbosity=opts.verbosity,
               calculators=calculators,
               files=args)
sys.exit(len(results.errors + results.failures))



from __future__ import print_function
"""This module defines an ASE interface to Turbomole

http://www.turbomole.com/
"""
import os
import sys

import numpy as np

from ase.units import Hartree, Bohr
from ase.io import read, write
from ase.calculators.general import Calculator


class Turbomole(Calculator):
    def __init__(self, label='turbomole',
                 calculate_energy='dscf', calculate_forces='grad',
                 post_HF=False):
        self.label = label
        self.converged = False
        
        # set calculators for energy and forces
        self.calculate_energy = calculate_energy
        self.calculate_forces = calculate_forces

        # turbomole has no stress
        self.stress = np.empty(6)
        
        # storage for energy and forces
        self.e_total = None
        self.forces = None
        self.updated = False
        
        # atoms must be set
        self.atoms = None
        
        # POST-HF method
        self.post_HF = post_HF

    def initialize(self, atoms):
        self.numbers = atoms.get_atomic_numbers().copy()
        self.species = []
        for a, Z in enumerate(self.numbers):
            self.species.append(Z)
        self.converged = False
        
    def execute(self, command):
        from subprocess import Popen, PIPE
        try:
            # the sub process gets started here
            proc = Popen([command], shell=True, stderr=PIPE)
            error = proc.communicate()[1]
            # check the error output
            if 'abnormally' in error:
                raise OSError(error)
            print('TM command: ', command, 'successfully executed')
        except OSError as e:
            print('Execution failed:', e, file=sys.stderr)
            sys.exit(1)

    def get_potential_energy(self, atoms):
        # update atoms
        self.updated = self.e_total is None
        self.set_atoms(atoms)
        # if update of energy is necessary
        if self.update_energy:
            # calculate energy
            self.execute(self.calculate_energy + ' > ASE.TM.energy.out')
            # check for convergence of dscf cycle
            if os.path.isfile('dscf_problem'):
                print('Turbomole scf energy calculation did not converge')
                raise RuntimeError(
                    'Please run Turbomole define and come thereafter back')
            # read energy
            self.read_energy()

        self.update_energy = False
        return self.e_total

    def get_forces(self, atoms):
        # update atoms
        self.updated = self.forces is None
        self.set_atoms(atoms)
        # complete energy calculations
        if self.update_energy:
            self.get_potential_energy(atoms)
        # if update of forces is necessary
        if self.update_forces:
            # calculate forces
            self.execute(self.calculate_forces + ' > ASE.TM.forces.out')
            # read forces
            self.read_forces()

        self.update_forces = False
        return self.forces.copy()
    
    def get_stress(self, atoms):
        return self.stress
        
    def set_atoms(self, atoms):
        if self.atoms == atoms:
            if (self.updated and os.path.isfile('coord')):
                self.updated = False
                a = read('coord').get_positions()
                if np.allclose(a, atoms.get_positions(), rtol=0, atol=1e-13):
                    return
            else:
                return
        # performs an update of the atoms
        write('coord', atoms)
        Calculator.set_atoms(self, atoms)
        # energy and forces must be re-calculated
        self.update_energy = True
        self.update_forces = True
        
    def read_energy(self):
        """Read Energy from Turbomole energy file."""
        text = open('energy', 'r').read().lower()
        lines = iter(text.split('\n'))

        # Energy:
        for line in lines:
            if line.startswith('$end'):
                break
            elif line.startswith('$'):
                pass
            else:
                energy_tmp = float(line.split()[1])
                if self.post_HF:
                    energy_tmp += float(line.split()[4])
        # update energy units
        self.e_total = energy_tmp * Hartree

    def read_forces(self):
        """Read Forces from Turbomole gradient file."""
        file = open('gradient', 'r')
        lines = file.readlines()
        file.close()

        forces = np.array([[0, 0, 0]])
        
        nline = len(lines)
        iline = -1
        
        for i in range(nline):
            if 'cycle' in lines[i]:
                iline = i
        
        if iline < 0:
            raise RuntimeError('Please check TURBOMOLE gradients')

        # next line
        iline += len(self.atoms) + 1
        # $end line
        nline -= 1
        # read gradients
        for i in range(iline, nline):
            line = lines[i].replace('D', 'E')
            tmp = np.array([[float(f) for f in line.split()[0:3]]])
            forces = np.concatenate((forces, tmp))
        # Note the '-' sign for turbomole, to get forces
        self.forces = (-np.delete(forces, np.s_[0:1], axis=0)) * Hartree / Bohr

    def calculation_required(self, atoms, properties):
        if self.atoms != atoms:
            return True
        for prop in properties:
            if prop == 'energy' and self.e_total is None:
                return True
            elif prop == 'forces' and self.forces is None:
                return True
        return False



"""
The following contains a database of small molecules

Data for the G2/97 database are from
Raghavachari, Redfern, and Pople, J. Chem. Phys. Vol. 106, 1063 (1997).
See http://www.cse.anl.gov/Catalysis_and_Energy_Conversion/Computational_Thermochemistry.shtml for the original files.

All numbers are experimental values, except for coordinates, which are
MP2(full)/6-31G(d) optimized geometries (from http://www.cse.anl.gov/OldCHMwebsiteContent/compmat/G2-97.htm)

Atomic species:
ref: Curtiss et al. JCP 106, 1063 (1997).
'Enthalpy' is the experimental enthalpies of formation at 0K
'thermal correction' is the thermal corrections H(298)-H(0)

Molecular species:
ref: Staroverov et al. JCP 119, 12129 (2003)
'Enthalpy' is the experimental enthalpies of formation at 298K
'ZPE' is the zero-point energies
'thermal correction' is the thermal enthalpy corrections H(298K) - H_exp(0K)
ZPE and thermal corrections are estimated from B3LYP geometries and vibrations.

For details about G2-1 and G2-2 sets see doi:10.1063/1.477422.

Experimental ionization potentials are from http://srdata.nist.gov/cccbdb/
Information presented on these pages is considered public information
and may be distributed or copied http://www.nist.gov/public_affairs/disclaimer.cfm

"""

from ase.atoms import string2symbols

atom_names = ['H','Li','Be','C','N','O','F','Na','Si','P','S','Cl']

molecule_names = ['LiH','BeH','CH','CH2_s3B1d','CH2_s1A1d','CH3','CH4','NH','NH2','NH3','OH','H2O','HF','SiH2_s1A1d','SiH2_s3B1d','SiH3','SiH4','PH2','PH3','SH2','HCl','Li2','LiF','C2H2','C2H4','C2H6','CN','HCN','CO','HCO','H2CO','CH3OH','N2','N2H4','NO','O2','H2O2','F2','CO2','Na2','Si2','P2','S2','Cl2','NaCl','SiO','CS','SO','ClO','ClF','Si2H6','CH3Cl','CH3SH','HOCl','SO2']

data = {
'H': {
    'name': 'Hydrogen',
    'database': 'G2-1',
    'symbols': 'H',
    'magmoms': [1.],
    'charges': None,
    'enthalpy': 51.63,
    'thermal correction': 1.01,
    'ionization energy': 13.60,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'Li': {
    'name': 'Lithium',
    'database': 'G2-1',
    'symbols': 'Li',
    'magmoms': [1.],
    'charges': None,
    'enthalpy': 37.69,
    'thermal correction': 1.10,
    'ionization energy': 5.39,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'Be': {
    'name': 'Beryllium',
    'database': 'G2-1',
    'symbols': 'Be',
    'magmoms': None,
    'charges': None,
    'enthalpy': 76.48,
    'thermal correction': 0.46,
    'ionization energy': 9.32,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'C': {
    'name': 'Carbon',
    'database': 'G2-1',
    'symbols': 'C',
    'magmoms': [2.],
    'charges': None,
    'enthalpy': 169.98,
    'thermal correction': 0.25,
    'ionization energy': 11.26,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'N': {
    'name': 'Nitrogen',
    'database': 'G2-1',
    'symbols': 'N',
    'magmoms': [3.],
    'charges': None,
    'enthalpy': 112.53,
    'thermal correction': 1.04,
    'ionization energy': 14.53,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'O': {
    'name': 'Oxygen',
    'database': 'G2-1',
    'symbols': 'O',
    'magmoms': [2.],
    'charges': None,
    'enthalpy': 58.99,
    'thermal correction': 1.04,
    'ionization energy': 13.62,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'F': {
    'name': 'Fluorine',
    'database': 'G2-1',
    'symbols': 'F',
    'magmoms': [1.],
    'charges': None,
    'enthalpy': 18.47,
    'thermal correction': 1.05,
    'ionization energy': 17.42,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'Na': {
    'name': 'Sodium',
    'database': 'G2-1',
    'symbols': 'Na',
    'magmoms': [1.],
    'charges': None,
    'enthalpy': 25.69,
    'thermal correction': 1.54,
    'ionization energy': 5.14,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'Si': {
    'name': 'Silicon',
    'database': 'G2-1',
    'symbols': 'Si',
    'magmoms': [2.],
    'charges': None,
    'enthalpy': 106.60,
    'thermal correction': 0.76,
    'ionization energy': 8.15,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'P': {
    'name': 'Phosphorus',
    'database': 'G2-1',
    'symbols': 'P',
    'magmoms': [3.],
    'charges': None,
    'enthalpy': 75.42,
    'thermal correction': 1.28,
    'ionization energy': 10.49,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'S': {
    'name': 'Sulfur',
    'database': 'G2-1',
    'symbols': 'S',
    'magmoms': [2.],
    'charges': None,
    'enthalpy': 65.66,
    'thermal correction': 1.05,
    'ionization energy': 10.36,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'Cl': {
    'name': 'Chlorine',
    'database': 'G2-1',
    'symbols': 'Cl',
    'magmoms': [1.],
    'charges': None,
    'enthalpy': 28.59,
    'thermal correction': 1.10,
    'ionization energy': 12.97,
    'positions': [[ 0.  ,  0.  ,  0.]],
    },
'LiH': {
    'description': "Lithium hydride (LiH), C*v symm.",
    'name': "LiH",
    'database': 'G2-1',
    'enthalpy': 33.3,
    'ZPE': 2.0149,
    'thermal correction': 2.0783,
    'ionization energy': 7.90,
    'symbols': 'LiH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.  ,  0.  ,  0.41],
                  [ 0.  ,  0.  , -1.23]]},
'BeH': {
    'description': "Beryllium hydride (BeH), D*h symm.",
    'name': "BeH",
    'database': 'G2-1',
    'enthalpy': 81.7,
    'ZPE': 2.9073,
    'thermal correction': 2.0739,
    'ionization energy': 8.21,
    'symbols': 'BeH',
    'magmoms': [ 0.8,  0.2],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.269654],
                  [ 0.      ,  0.      , -1.078616]]},
'CH': {
    'description': "CH radical. Doublet, C*v symm.",
    'name': "CH",
    'database': 'G2-1',
    'enthalpy': 142.5,
    'ZPE': 3.9659,
    'thermal correction': 2.0739,
    'ionization energy': 10.64,
    'symbols': 'CH',
    'magmoms': [ 1.,  0.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.160074],
                  [ 0.      ,  0.      , -0.960446]]},
'CH2_s3B1d': {
    'description': "Triplet methylene (CH2), C2v symm, 3-B1.",
    'name': "CH_2 (^3B_1)",
    'database': 'G2-1',
    'enthalpy': 93.7,
    'ZPE': 10.6953,
    'thermal correction': 2.3877,
    'ionization energy': 10.40,
    'symbols': 'CHH',
    'magmoms': [ 2.,  0.,  0.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.110381],
                  [ 0.      ,  0.982622, -0.331142],
                  [ 0.      , -0.982622, -0.331142]]},
'CH2_s1A1d': {
    'description': "Singlet methylene (CH2), C2v symm, 1-A1.",
    'name': "CH_2 (^1A_1)",
    'database': 'G2-1',
    'enthalpy': 102.8,
    'ZPE': 10.2422,
    'thermal correction': 2.3745,
    'symbols': 'CHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.174343],
                  [ 0.      ,  0.862232, -0.523029],
                  [ 0.      , -0.862232, -0.523029]]},
'CH3': {
    'description': "Methyl radical (CH3), D3h symm.",
    'name': "CH_3",
    'database': 'G2-1',
    'enthalpy': 35.0,
    'ZPE': 18.3383,
    'thermal correction': 2.5383,
    'ionization energy': 9.84,
    'symbols': 'CHHH',
    'magmoms': [ 1.,  0.,  0.,  0.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.      ],
                  [ 0.      ,  1.07841 ,  0.      ],
                  [ 0.93393 , -0.539205,  0.      ],
                  [-0.93393 , -0.539205,  0.      ]]},
'CH4': {
    'description': "Methane (CH4), Td symm.",
    'name': "CH_4",
    'database': 'G2-1',
    'enthalpy': -17.9,
    'ZPE': 27.6744,
    'thermal correction': 2.3939,
    'ionization energy': 12.64,
    'vertical ionization energy': 13.60,
    'symbols': 'CHHHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.      ],
                  [ 0.629118,  0.629118,  0.629118],
                  [-0.629118, -0.629118,  0.629118],
                  [ 0.629118, -0.629118, -0.629118],
                  [-0.629118,  0.629118, -0.629118]]},
'NH': {
    'description': "NH, triplet, C*v symm.",
    'name': "NH",
    'database': 'G2-1',
    'enthalpy': 85.2,
    'ZPE': 4.5739,
    'thermal correction': 2.0739,
    'ionization energy': 13.10,
    'vertical ionization energy': 13.49,
    'symbols': 'NH',
    'magmoms': [ 2.,  0.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.129929],
                  [ 0.      ,  0.      , -0.909501]]},
'NH2': {
    'description': "NH2 radical, C2v symm, 2-B1.",
    'name': "NH_2",
    'database': 'G2-1',
    'enthalpy': 45.1,
    'ZPE': 11.7420,
    'thermal correction': 2.3726,
    'ionization energy': 10.78,
    'vertical ionization energy': 12.00,
    'symbols': 'NHH',
    'magmoms': [ 1.,  0.,  0.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.14169 ],
                  [ 0.      ,  0.806442, -0.495913],
                  [ 0.      , -0.806442, -0.495913]]},
'NH3': {
    'description': "Ammonia (NH3), C3v symm.",
    'name': "NH_3",
    'database': 'G2-1',
    'enthalpy': -11.0,
    'ZPE': 21.2462,
    'thermal correction': 2.3896,
    'ionization energy': 10.07,
    'vertical ionization energy': 10.82,
    'symbols': 'NHHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.116489],
                  [ 0.      ,  0.939731, -0.271808],
                  [ 0.813831, -0.469865, -0.271808],
                  [-0.813831, -0.469865, -0.271808]]},
'OH': {
    'description': "OH radical, C*v symm.",
    'name': "OH",
    'database': 'G2-1',
    'enthalpy': 9.4,
    'ZPE': 5.2039,
    'thermal correction': 2.0739,
    'ionization energy': 13.02,
    'symbols': 'OH',
    'magmoms': [ 0.5,  0.5],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.108786],
                  [ 0.      ,  0.      , -0.870284]]},
'H2O': {
    'description': "Water (H2O), C2v symm.",
    'name': "H_2O",
    'database': 'G2-1',
    'enthalpy': -57.8,
    'ZPE': 13.2179,
    'thermal correction': 2.3720,
    'ionization energy': 12.62,
    'symbols': 'OHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.119262],
                  [ 0.      ,  0.763239, -0.477047],
                  [ 0.      , -0.763239, -0.477047]]},
'HF': {
    'description': "Hydrogen fluoride (HF), C*v symm.",
    'name': "HF",
    'database': 'G2-1',
    'enthalpy': -65.1,
    'ZPE': 5.7994,
    'thermal correction': 2.0733,
    'ionization energy': 16.03,
    'vertical ionization energy': 16.12,
    'symbols': 'FH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.093389],
                  [ 0.      ,  0.      , -0.840502]]},
'SiH2_s1A1d': {
    'description': "Singlet silylene (SiH2), C2v symm, 1-A1.",
    'name': "SiH_2 (^1A_1)",
    'database': 'G2-1',
    'enthalpy': 65.2,
    'ZPE': 7.1875,
    'thermal correction': 2.3927,
    'ionization energy': 8.92,
    'symbols': 'SiHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.131272],
                  [ 0.      ,  1.096938, -0.918905],
                  [ 0.      , -1.096938, -0.918905]]},
'SiH2_s3B1d': {
    'description': "Triplet silylene (SiH2), C2v symm, 3-B1.",
    'name': "SiH_2 (^3B_1)",
    'database': 'G2-1',
    'enthalpy': 86.2,
    'ZPE': 7.4203,
    'thermal correction': 2.4078,
    'symbols': 'SiHH',
    'magmoms': [ 2.,  0.,  0.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.094869],
                  [ 0.      ,  1.271862, -0.664083],
                  [ 0.      , -1.271862, -0.664083]]},
'SiH3': {
    'description': "Silyl radical (SiH3), C3v symm.",
    'name': "SiH_3",
    'database': 'G2-1',
    'enthalpy': 47.9,
    'ZPE': 13.0898,
    'thermal correction': 2.4912,
    'ionization energy': 8.14,
    'vertical ionization energy': 8.74,
    'symbols': 'SiHHH',
    'magmoms': [ 1.,  0.,  0.,  0.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.079299],
                  [ 0.      ,  1.41328 , -0.370061],
                  [ 1.223937, -0.70664 , -0.370061],
                  [-1.223937, -0.70664 , -0.370061]]},
'SiH4': {
    'description': "Silane (SiH4), Td symm.",
    'name': "SiH_4",
    'database': 'G2-1',
    'enthalpy': 8.2,
    'ZPE': 19.2664,
    'thermal correction': 2.5232,
    'ionization energy': 11.00,
    'vertical ionization energy': 12.30,
    'symbols': 'SiHHHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.      ],
                  [ 0.856135,  0.856135,  0.856135],
                  [-0.856135, -0.856135,  0.856135],
                  [-0.856135,  0.856135, -0.856135],
                  [ 0.856135, -0.856135, -0.856135]]},
'PH2': {
    'description': "PH2 radical, C2v symm.",
    'name': "PH_2",
    'database': 'G2-1',
    'enthalpy': 33.1,
    'ZPE': 8.2725,
    'thermal correction': 2.3845,
    'ionization energy': 9.82,
    'symbols': 'PHH',
    'magmoms': [ 1.,  0.,  0.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.115396],
                  [ 0.      ,  1.025642, -0.865468],
                  [ 0.      , -1.025642, -0.865468]]},
'PH3': {
    'description': "Phosphine (PH3), C3v symm.",
    'name': "PH_3",
    'database': 'G2-1',
    'enthalpy': 1.3,
    'ZPE': 14.7885,
    'thermal correction': 2.4203,
    'ionization energy': 9.87,
    'vertical ionization energy': 10.95,
    'symbols': 'PHHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.124619],
                  [ 0.      ,  1.200647, -0.623095],
                  [ 1.039791, -0.600323, -0.623095],
                  [-1.039791, -0.600323, -0.623095]]},
'SH2': {
    'description': "Hydrogen sulfide (H2S), C2v symm.",
    'name': "SH_2",
    'database': 'G2-1',
    'enthalpy': -4.9,
    'ZPE': 9.3129,
    'thermal correction': 2.3808,
    'ionization energy': 10.46,
    'vertical ionization energy': 10.50,
    'symbols': 'SHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.102135],
                  [ 0.      ,  0.974269, -0.817083],
                  [ 0.      , -0.974269, -0.817083]]},
'HCl': {
    'description': "Hydrogen chloride (HCl), C*v symm.",
    'name': "HCl",
    'database': 'G2-1',
    'enthalpy': -22.1,
    'ZPE': 4.1673,
    'thermal correction': 2.0739,
    'ionization energy': 12.74,
    'symbols': 'ClH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.07111 ],
                  [ 0.      ,  0.      , -1.208868]]},
'Li2': {
    'description': "Dilithium (Li2), D*h symm.",
    'name': "Li_2",
    'database': 'G2-1',
    'enthalpy': 51.6,
    'ZPE': 0.4838,
    'thermal correction': 2.3086,
    'ionization energy': 5.11,
    'symbols': 'LiLi',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.     ,  0.     ,  1.38653],
                  [ 0.     ,  0.     , -1.38653]]},
'LiF': {
    'description': "Lithium Fluoride (LiF), C*v symm.",
    'name': "LiF",
    'database': 'G2-1',
    'enthalpy': -80.1,
    'ZPE': 1.4019,
    'thermal correction': 2.0990,
    'ionization energy': 11.30,
    'symbols': 'LiF',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      , -1.174965],
                  [ 0.      ,  0.      ,  0.391655]]},
'C2H2': {
    'description': "Acetylene (C2H2), D*h symm.",
    'name': "C_2H_2",
    'database': 'G2-1',
    'enthalpy': 54.2,
    'ZPE': 16.6001,
    'thermal correction': 2.4228,
    'ionization energy': 11.40,
    'vertical ionization energy': 11.49,
    'symbols': 'CCHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.     ,  0.     ,  0.60808],
                  [ 0.     ,  0.     , -0.60808],
                  [ 0.     ,  0.     , -1.67399],
                  [ 0.     ,  0.     ,  1.67399]]},
'C2H4': {
    'description': "Ethylene (H2C=CH2), D2h symm.",
    'name': "C_2H_4",
    'database': 'G2-1',
    'enthalpy': 12.5,
    'ZPE': 31.5267,
    'thermal correction': 2.5100,
    'ionization energy': 11.40,
    'vertical ionization energy': 11.49,
    'symbols': 'CCHHHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.66748 ],
                  [ 0.      ,  0.      , -0.66748 ],
                  [ 0.      ,  0.922832,  1.237695],
                  [ 0.      , -0.922832,  1.237695],
                  [ 0.      ,  0.922832, -1.237695],
                  [ 0.      , -0.922832, -1.237695]]},
'C2H6': {
    'description': "Ethane (H3C-CH3), D3d symm.",
    'name': "C_2H_6",
    'database': 'G2-1',
    'enthalpy': -20.1,
    'ZPE': 46.0950,
    'thermal correction': 2.7912,
    'symbols': 'CCHHHHHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.762209],
                  [ 0.      ,  0.      , -0.762209],
                  [ 0.      ,  1.018957,  1.157229],
                  [-0.882443, -0.509479,  1.157229],
                  [ 0.882443, -0.509479,  1.157229],
                  [ 0.      , -1.018957, -1.157229],
                  [-0.882443,  0.509479, -1.157229],
                  [ 0.882443,  0.509479, -1.157229]]},
'CN': {
    'description': "Cyano radical (CN), C*v symm, 2-Sigma+.",
    'name': "CN",
    'database': 'G2-1',
    'enthalpy': 104.9,
    'ZPE': 3.0183,
    'thermal correction': 2.0739,
    'ionization energy': 13.60,
    'symbols': 'CN',
    'magmoms': [ 1.,  0.],
    'charges': None,
    'positions': [[ 0.      ,  0.      , -0.611046],
                  [ 0.      ,  0.      ,  0.523753]]},
'HCN': {
    'description': "Hydrogen cyanide (HCN), C*v symm.",
    'name': "HCN",
    'database': 'G2-1',
    'enthalpy': 31.5,
    'ZPE': 10.2654,
    'thermal correction': 2.1768,
    'ionization energy': 13.60,
    'vertical ionization energy': 13.61,
    'symbols': 'CNH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      , -0.511747],
                  [ 0.      ,  0.      ,  0.664461],
                  [ 0.      ,  0.      , -1.580746]]},
'CO': {
    'description': "Carbon monoxide (CO), C*v symm.",
    'name': "CO",
    'database': 'G2-1',
    'enthalpy': -26.4,
    'ZPE': 3.1062,
    'thermal correction': 2.0739,
    'ionization energy': 14.01,
    'vertical ionization energy': 14.01,
    'symbols': 'OC',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.493003],
                  [ 0.      ,  0.      , -0.657337]]},
'HCO': {
    'description': "HCO radical, Bent Cs symm.",
    'name': "HCO",
    'database': 'G2-1',
    'enthalpy': 10.0,
    'ZPE': 8.0290,
    'thermal correction': 2.3864,
    'ionization energy': 8.12,
    'vertical ionization energy': 9.31,
    'symbols': 'COH',
    'magmoms': [ 1.,  0.,  0.],
    'charges': None,
    'positions': [[ 0.06256 ,  0.593926,  0.      ],
                  [ 0.06256 , -0.596914,  0.      ],
                  [-0.875835,  1.211755,  0.      ]]},
'H2CO': {
    'description': "Formaldehyde (H2C=O), C2v symm.",
    'name': "H_2CO",
    'database': 'G2-1',
    'enthalpy': -26.0,
    'ZPE': 16.4502,
    'thermal correction': 2.3927,
    'ionization energy': 10.88,
    'vertical ionization energy': 10.88,
    'symbols': 'OCHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.683501],
                  [ 0.      ,  0.      , -0.536614],
                  [ 0.      ,  0.93439 , -1.124164],
                  [ 0.      , -0.93439 , -1.124164]]},
'CH3OH': {
    'description': "Methanol (CH3-OH), Cs symm.",
    'name': "H_3COH",
    'database': 'G2-1',
    'enthalpy': -48.0,
    'ZPE': 31.6635,
    'thermal correction': 2.6832,
    'ionization energy': 10.84,
    'vertical ionization energy': 10.96,
    'symbols': 'COHHHH',
    'magmoms': None,
    'charges': None,
    'positions': [[-0.047131,  0.664389,  0.      ],
                  [-0.047131, -0.758551,  0.      ],
                  [-1.092995,  0.969785,  0.      ],
                  [ 0.878534, -1.048458,  0.      ],
                  [ 0.437145,  1.080376,  0.891772],
                  [ 0.437145,  1.080376, -0.891772]]},
'N2': {
    'description': "N2 molecule, D*h symm.",
    'name': "N_2",
    'database': 'G2-1',
    'enthalpy': 0.0,
    'ZPE': 3.4243,
    'thermal correction': 2.0733,
    'ionization energy': 15.58,
    'vertical ionization energy': 15.58,
    'symbols': 'NN',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.     ,  0.     ,  0.56499],
                  [ 0.     ,  0.     , -0.56499]]},
'N2H4': {
    'description': "Hydrazine (H2N-NH2), C2 symm.",
    'name': "H_2NNH_2",
    'database': 'G2-1',
    'enthalpy': 22.8,
    'ZPE': 32.9706,
    'thermal correction': 2.6531,
    'ionization energy': 8.10,
    'vertical ionization energy': 8.98,
    'symbols': 'NNHHHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.718959, -0.077687],
                  [ 0.      , -0.718959, -0.077687],
                  [ 0.211082,  1.092752,  0.847887],
                  [-0.948214,  1.005026, -0.304078],
                  [-0.211082, -1.092752,  0.847887],
                  [ 0.948214, -1.005026, -0.304078]]},
'NO': {
    'description': "NO radical, C*v symm, 2-Pi.",
    'name': "NO",
    'database': 'G2-1',
    'enthalpy': 21.6,
    'ZPE': 2.7974,
    'thermal correction': 2.0745,
    'ionization energy': 9.26,
    'vertical ionization energy': 9.26,
    'symbols': 'NO',
    'magmoms': [ 0.6,  0.4],
    'charges': None,
    'positions': [[ 0.      ,  0.      , -0.609442],
                  [ 0.      ,  0.      ,  0.533261]]},
'O2': {
    'description': "O2 molecule, D*h symm, Triplet.",
    'name': "O_2",
    'database': 'G2-1',
    'enthalpy': 0.0,
    'ZPE': 2.3444,
    'thermal correction': 2.0752,
    'ionization energy': 12.07,
    'vertical ionization energy': 12.30,
    'symbols': 'OO',
    'magmoms': [ 1.,  1.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.622978],
                  [ 0.      ,  0.      , -0.622978]]},
'H2O2': {
    'description': "Hydrogen peroxide (HO-OH), C2 symm.",
    'name': "HOOH",
    'database': 'G2-1',
    'enthalpy': -32.5,
    'ZPE': 16.4081,
    'thermal correction': 2.6230,
    'ionization energy': 10.58,
    'vertical ionization energy': 11.70,
    'symbols': 'OOHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.734058, -0.05275 ],
                  [ 0.      , -0.734058, -0.05275 ],
                  [ 0.839547,  0.880752,  0.422001],
                  [-0.839547, -0.880752,  0.422001]]},
'F2': {
    'description': "F2 molecule, D*h symm.",
    'name': "F_2",
    'database': 'G2-1',
    'enthalpy': 0.0,
    'ZPE': 1.5179,
    'thermal correction': 2.0915,
    'ionization energy': 15.70,
    'vertical ionization energy': 15.70,
    'symbols': 'FF',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.710304],
                  [ 0.      ,  0.      , -0.710304]]},
'CO2': {
    'description': "Carbon dioxide (CO2), D*h symm.",
    'name': "CO_2",
    'database': 'G2-1',
    'enthalpy': -94.1,
    'ZPE': 7.3130,
    'thermal correction': 2.2321,
    'ionization energy': 13.78,
    'vertical ionization energy': 13.78,
    'symbols': 'COO',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.      ],
                  [ 0.      ,  0.      ,  1.178658],
                  [ 0.      ,  0.      , -1.178658]]},
'Na2': {
    'description': "Disodium (Na2), D*h symm.",
    'name': "Na_2",
    'database': 'G2-1',
    'enthalpy': 34.0,
    'ZPE': 0.2246,
    'thermal correction': 2.4699,
    'ionization energy': 4.89,
    'symbols': 'NaNa',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  1.576262],
                  [ 0.      ,  0.      , -1.576262]]},
'Si2': {
    'description': "Si2 molecule, D*h symm, Triplet (3-Sigma-G-).",
    'name': "Si_2",
    'database': 'G2-1',
    'enthalpy': 139.9,
    'ZPE': 0.7028,
    'thermal correction': 2.2182,
    'ionization energy': 7.90,
    'symbols': 'SiSi',
    'magmoms': [ 1.,  1.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  1.130054],
                  [ 0.      ,  0.      , -1.130054]]},
'P2': {
    'description': "P2 molecule, D*h symm.",
    'name': "P_2",
    'database': 'G2-1',
    'enthalpy': 34.3,
    'ZPE': 1.1358,
    'thermal correction': 2.1235,
    'ionization energy': 10.53,
    'vertical ionization energy': 10.62,
    'symbols': 'PP',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.966144],
                  [ 0.      ,  0.      , -0.966144]]},
'S2': {
    'description': "S2 molecule, D*h symm, triplet.",
    'name': "S_2",
    'database': 'G2-1',
    'enthalpy': 30.7,
    'ZPE': 1.0078,
    'thermal correction': 2.1436,
    'ionization energy': 9.36,
    'vertical ionization energy': 9.55,
    'symbols': 'SS',
    'magmoms': [ 1.,  1.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.960113],
                  [ 0.      ,  0.      , -0.960113]]},
'Cl2': {
    'description': "Cl2 molecule, D*h symm.",
    'name': "Cl_2",
    'database': 'G2-1',
    'enthalpy': 0.0,
    'ZPE': 0.7737,
    'thermal correction': 2.1963,
    'ionization energy': 11.48,
    'vertical ionization energy': 11.49,
    'symbols': 'ClCl',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  1.007541],
                  [ 0.      ,  0.      , -1.007541]]},
'NaCl': {
    'description': "Sodium Chloride (NaCl), C*v symm.",
    'name': "NaCl",
    'database': 'G2-1',
    'enthalpy': -43.6,
    'ZPE': 0.5152,
    'thermal correction': 2.2935,
    'ionization energy': 9.20,
    'vertical ionization energy': 9.80,
    'symbols': 'NaCl',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.     ,  0.     , -1.45166],
                  [ 0.     ,  0.     ,  0.93931]]},
'SiO': {
    'description': "Silicon monoxide (SiO), C*v symm.",
    'name': "SiO",
    'database': 'G2-1',
    'enthalpy': -24.6,
    'ZPE': 1.7859,
    'thermal correction': 2.0821,
    'ionization energy': 11.49,
    'symbols': 'SiO',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.560846],
                  [ 0.      ,  0.      , -0.98148 ]]},
'CS': {
    'description': "Carbon monosulfide (CS), C*v symm.",
    'name': "SC",
    'database': 'G2-1',
    'enthalpy': 66.9,
    'ZPE': 1.8242,
    'thermal correction': 2.0814,
    'ionization energy': 11.33,
    'symbols': 'CS',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      , -1.123382],
                  [ 0.      ,  0.      ,  0.421268]]},
'SO': {
    'description': "Sulfur monoxide (SO), C*v symm, triplet.",
    'name': "SO",
    'database': 'G2-1',
    'enthalpy': 1.2,
    'ZPE': 1.6158,
    'thermal correction': 2.0877,
    'ionization energy': 11.29,
    'symbols': 'OS',
    'magmoms': [ 1.,  1.],
    'charges': None,
    'positions': [[ 0.      ,  0.      , -1.015992],
                  [ 0.      ,  0.      ,  0.507996]]},
'ClO': {
    'description': "ClO radical, C*v symm, 2-PI.",
    'name': "ClO",
    'database': 'G2-1',
    'enthalpy': 24.2,
    'ZPE': 1.1923,
    'thermal correction': 2.1172,
    'ionization energy': 10.89,
    'vertical ionization energy': 11.01,
    'symbols': 'ClO',
    'magmoms': [ 1.,  0.],
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.514172],
                  [ 0.      ,  0.      , -1.092615]]},
'ClF': {
    'description': "ClF molecule, C*v symm, 1-SG.",
    'name': "FCl",
    'database': 'G2-1',
    'enthalpy': -13.2,
    'ZPE': 1.1113,
    'thermal correction': 2.1273,
    'ionization energy': 12.66,
    'vertical ionization energy': 12.77,
    'symbols': 'FCl',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      , -1.084794],
                  [ 0.      ,  0.      ,  0.574302]]},
'Si2H6': {
    'description': "Disilane (H3Si-SiH3), D3d symm.",
    'name': "Si_2H_6",
    'database': 'G2-1',
    'enthalpy': 19.1,
    'ZPE': 30.2265,
    'thermal correction': 3.7927,
    'ionization energy': 9.74,
    'vertical ionization energy': 10.53,
    'symbols': 'SiSiHHHHHH',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  1.167683],
                  [ 0.      ,  0.      , -1.167683],
                  [ 0.      ,  1.393286,  1.68602 ],
                  [-1.206621, -0.696643,  1.68602 ],
                  [ 1.206621, -0.696643,  1.68602 ],
                  [ 0.      , -1.393286, -1.68602 ],
                  [-1.206621,  0.696643, -1.68602 ],
                  [ 1.206621,  0.696643, -1.68602 ]]},
'CH3Cl': {
    'description': "Methyl chloride (CH3Cl), C3v symm.",
    'name': "CH_3Cl",
    'database': 'G2-1',
    'enthalpy': -19.6,
    'ZPE': 23.3013,
    'thermal correction': 2.4956,
    'symbols': 'CClHHH',
    'ionization energy': 11.26,
    'vertical ionization energy': 11.29,
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      , -1.121389],
                  [ 0.      ,  0.      ,  0.655951],
                  [ 0.      ,  1.029318, -1.47428 ],
                  [ 0.891415, -0.514659, -1.47428 ],
                  [-0.891415, -0.514659, -1.47428 ]]},
'CH3SH': {
    'description': "Methanethiol (H3C-SH), Staggered, Cs symm.",
    'name': "H_3CSH",
    'database': 'G2-1',
    'enthalpy': -5.5,
    'ZPE': 28.3973,
    'thermal correction': 2.8690,
    'ionization energy': 9.44,
    'vertical ionization energy': 9.44,
    'symbols': 'CSHHHH',
    'magmoms': None,
    'charges': None,
    'positions': [[-0.047953,  1.149519,  0.      ],
                  [-0.047953, -0.664856,  0.      ],
                  [ 1.283076, -0.823249,  0.      ],
                  [-1.092601,  1.461428,  0.      ],
                  [ 0.432249,  1.551207,  0.892259],
                  [ 0.432249,  1.551207, -0.892259]]},
'HOCl': {
    'description': "HOCl molecule, Cs symm.",
    'name': "HOCl",
    'database': 'G2-1',
    'enthalpy': -17.8,
    'ZPE': 8.1539,
    'thermal correction': 2.4416,
    'ionization energy': 11.12,
    'symbols': 'OHCl',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.036702,  1.113517,  0.      ],
                  [-0.917548,  1.328879,  0.      ],
                  [ 0.036702, -0.602177,  0.      ]]},
'SO2': {
    'description': "Sulfur dioxide (SO2), C2v symm.",
    'name': "SO_2",
    'database': 'G2-1',
    'enthalpy': -71.0,
    'ZPE': 4.3242,
    'thermal correction': 2.5245,
    'ionization energy': 12.35,
    'vertical ionization energy': 12.50,
    'symbols': 'SOO',
    'magmoms': None,
    'charges': None,
    'positions': [[ 0.      ,  0.      ,  0.370268],
                  [ 0.      ,  1.277617, -0.370268],
                  [ 0.      , -1.277617, -0.370268]]},
}

def get_ionization_energy(name, vertical=True):
    """Return the experimental ionization energy from the database.

    If vertical is True, the vertical ionization energy is returned if
    available.
    """
    if name not in data:
        raise KeyError('System %s not in database.' % name)
    elif 'ionization energy' not in data[name]:
        raise KeyError('No data on ionization energy for system %s.' % name)
    else:
        if vertical and 'vertical ionization energy' in data[name]:
            return data[name]['vertical ionization energy']
        else:
            return data[name]['ionization energy']


def get_atomization_energy(name):
    """Determine extrapolated experimental atomization energy from the database.

    The atomization energy is extrapolated from experimental heats of
    formation at room temperature, using calculated zero-point energies
    and thermal corrections.

    The atomization energy is returned in kcal/mol = 43.36 meV:

    >>> from ase.units import *; print kcal / mol
    0.0433641146392

    """
    d = data[name]
    e = d['enthalpy']
    z = d['ZPE']
    dh = d['thermal correction']
    ae = -e + z + dh
    for a in string2symbols(d['symbols']):
        h = data[a]['enthalpy']
        dh = data[a]['thermal correction']
        ae += h - dh
    return ae



from ase.cluster import Icosahedron
from ase.ga.particle_crossovers import CutSpliceCrossover
from random import shuffle

ico1 = Icosahedron('Cu', 3)
ico1.info['confid'] = 1
ico2 = Icosahedron('Ni', 3)
ico2.info['confid'] = 2

# TODO: Change this crossover to one for fixed particles
# op = CutSpliceCrossover({(28, 29): 2.0, (28, 28): 2.0, (29, 29): 2.0},
#                         keep_composition=False)
# a3, desc = op.get_new_individual([ico1, ico2])

# assert len(set(a3.get_chemical_symbols())) == 2
# assert len(a3) == 55

ico1.numbers[:20] = [28] * 20
shuffle(ico1.numbers)
ico2.numbers[:35] = [29] * 35
shuffle(ico2.numbers)
op = CutSpliceCrossover({(28, 29): 2.0, (28, 28): 2.0, (29, 29): 2.0})
a3, desc = op.get_new_individual([ico1, ico2])

assert a3.get_chemical_formula() == 'Cu35Ni20'

from ase.ga.particle_mutations import COM2surfPermutation
# from ase.ga.particle_mutations import RandomPermutation
# from ase.ga.particle_mutations import Poor2richPermutation
# from ase.ga.particle_mutations import Rich2poorPermutation

op = COM2surfPermutation(min_ratio=0.05)
a3, desc = op.get_new_individual([ico1])
a3.info['confid'] = 3

assert a3.get_chemical_formula() == 'Cu35Ni20'

aconf = op.get_atomic_configuration(a3)
core = aconf[1]
shell = aconf[-1]
for i, sym in zip(core, 6 * ['Ni'] + 6 * ['Cu']):
    a3[i].symbol = sym
for i, sym in zip(shell, 6 * ['Ni'] + 6 * ['Cu']):
    a3[i].symbol = sym

atomic_conf = op.get_atomic_configuration(a3, elements=['Cu'])[-2:]
cu3 = len([item for sublist in atomic_conf for item in sublist])
a4, desc = op.get_new_individual([a3])
atomic_conf = op.get_atomic_configuration(a4, elements=['Cu'])[-2:]
cu4 = len([item for sublist in atomic_conf for item in sublist])

assert abs(cu4 - cu3) == 1
